// -*- c++ -*-
// $Id$
/*
 * Copyright 2005-2007 Spencer Olson
 *
 * $Log$
 *
 */

/** \file
 * A uniform box with periodic (3D) boundary conditions.  Not too exciting.
 * This uniform box does test the teleporter boundary objects in creating
 * periodic boundary conditions.
 */

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <cfloat>

#include <olson-tools/strutil.h>
#include <olson-tools/physical.h>
#include <olson-tools/bfield.h>
#include <olson-tools/random/random.h>
#include <olson-tools/listutil.h>
#include <olson-tools/KeyedBin.h>
#include <olson-tools/GenericBin.h>

#include <dsmc/DSMC.h>
#include <dsmc/ParticleSrc.h>
#include <dsmc/Wall.h>

#include <stdexcept>
#include <sys/times.h>
#include <unistd.h>
#include <signal.h> // so we can issue ^C to gently stop the loading

static bool quit_requested = false;
void quitRequested(int inSignal) {
    quit_requested = true;
}

using namespace physical::constants;
using namespace physical::units;

const double Len = 1.0*m; /* to give similar results as the Argon case. */
const double temperature = 273.*K;

#define DIMENSIONS 1

const double Volume =
#if defined(DIMENSIONS)
#  if (DIMENSIONS == 3)
    CUBE(Len)
#  elif  (DIMENSIONS == 2)
    SQR(Len)
#  else
    Len
#  endif
#else
    CUBE(Len)
#endif
    ;

typedef octree::TeleporterBoundaryObject<Particle> tele_type;
const int num_tele = 6;
tele_type tele[num_tele] = {
    /*        normal     point             side-length  scaling     teleport-distance */
    tele_type(V3(-1,0,0), V3( 0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm, V3(-0.9999*Len,0,0)),
    tele_type(V3( 1,0,0), V3(-0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm, V3( 0.9999*Len,0,0)),
    tele_type(V3(0,-1,0), V3(0, 0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm, V3(0,-0.9999*Len,0)),
    tele_type(V3(0, 1,0), V3(0,-0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm, V3(0, 0.9999*Len,0)),
    tele_type(V3(0,0,-1), V3(0,0, 0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm, V3(0,0,-0.9999*Len)),
    tele_type(V3(0,0, 1), V3(0,0,-0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm, V3(0,0, 0.9999*Len))
};

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore);

GenericBin<double,1000> vbin[3] = {
    GenericBin<double,1000>(-2000*m/s,2000*m/s),
    GenericBin<double,1000>(-2000*m/s,2000*m/s),
    GenericBin<double,1000>(-2000*m/s,2000*m/s)
};

GenericBin<double,1000> tmpvbin[3] = {
    GenericBin<double,1000>(-2000*m/s,2000*m/s),
    GenericBin<double,1000>(-2000*m/s,2000*m/s),
    GenericBin<double,1000>(-2000*m/s,2000*m/s)
};

/** Truncated Gaussian distribution.
*/
class TruncatedGaussianDistrib {
  public:
    /** Contructor for gaussian distribution.
     * @param b
     *    Set \f$ \beta \f$ for the distribution defined by 
     *    \f$ {\rm e}^{-\beta v^{2}} \f$.
     * @param offs
     *    the center of the Gaussian distribution.
     */
    inline TruncatedGaussianDistrib(const double & b, const double & offs = 0.0) :
        beta(b), v0(offs) {}
    double beta;
    double v0;

    /** Evaluate \f$ {\rm e}^{-\beta v^{2}} \f$.
     * @param v
     *     Independent variable.
     */
    inline double distrib (const double & v) const {
        double eta = SQR(v-v0) * beta;
        static const double eta_max  = 0.9; /* trucation done at ~80% in distro tail */
        if (eta > eta_max) {
            return 0.0; /* truncation */
        } else {
            return exp(-eta);
        }
    }
};

class TruncatedGaussian3Distros {
  public:
    typedef TruncatedGaussianDistrib P_vx;
    typedef TruncatedGaussianDistrib P_vy;
    typedef TruncatedGaussianDistrib P_vz;
};




int main() {
    /* need to be able to add particles to the system. */
    int8_t particle_type = getParticleTypeIndex("Ar");

    const double Kn = 0.1;
    /* We'll ensure that dt ~= min(0.1 Coll.Time,0.1 1m/v_avg)  */
    static double dt_max =
        0.1 /* we need to allow excursions 1/10 the size of L */
        * Len /* characteristic length. */
        / sqrt( 8.0*K_B * temperature / pi / particle_info[particle_type].mass);
    const double dt_Kn =
        0.1 /* we need to allow excursions 1/10 the size of L */
        * (Kn * Len) /* mean free path. */
        / sqrt( 8.0*K_B * temperature / pi / particle_info[particle_type].mass);
    const double dt = std::min(dt_Kn, dt_max);
    static double t_max = 15.*ms + dt /* dt * 200.0 */;
    static int spit_out_data = 0;

    const double desired_samples = 115;
    const double sample_wait_time = 100.0*s; /* time before real sampling happens. */
    static int sample_interval = int( (t_max/*-sample_wait_time*/) / ( desired_samples * dt ) );
    static int write_interval = 1 * sample_interval;
    int write_particles_interval = write_interval; /* must be multiple of write_interval */


    ParticleSrc< FlatParticleDistrib, TruncatedGaussian3Distros > atomsrc;
    atomsrc.seed.type = particle_type;
    atomsrc.seed.dt   =  10*us;

    atomsrc.pos_distrib.delta_x         = 0.4999*Len;
#if defined(DIMENSIONS) && (DIMENSIONS < 3)
    atomsrc.pos_distrib.delta_x[Y]      = 0.0;
#  if  (DIMENSIONS < 2)
    atomsrc.pos_distrib.delta_x[Z]      = 0.0;
#  endif
#endif

    const double n0                     = 1.0 / (   M_SQRT2
                                                  * particle_info[particle_type].cross_section
                                                  * (Kn * Len)
                                                );
    atomsrc.flux                        = (n0 * Volume ) / dt;

    particle_info[particle_type].size   = atomsrc.flux * 1e-4 * dt;
    atomsrc.pulsed_source               = true;
    atomsrc.pulse_duration              = dt;
    atomsrc.origin                      = 0.0;
    atomsrc.temperature                 = temperature;
    atomsrc.velocity                    = 0.0;
    atomsrc.initialize();


    std::cout << "(" << getpid() << ")  t_max address:  " << (&t_max) << std::endl;
    std::cout << "(" << getpid() << ")  spit_out_data address:  " << (&spit_out_data) << std::endl;
    std::cout << "(" << getpid() << ")  sample_interval address:  " << (&sample_interval) << std::endl;
    std::cout << "(" << getpid() << ")  write_interval address:  " << (&write_interval) << std::endl;
    std::cout << "(" << getpid() << ")  n0:  " << n0 << std::endl;


    DSMC< RK2Integrator, octree::CMAndVolNotShrink > dsmc;
    dsmc.integrator.derivs = flatderivs;
    dsmc.allow_collisions = true;

    for (int i = 0; i < num_tele; i++) {
        dsmc.addBoundaryObject(&tele[i]);
    }

    dsmc.data_radius_of_influence = 4;

    /* ** plotting all particles. */
    std::ofstream fout("particles.plt");
    std::ofstream fdata("data.plt");
    std::ofstream fpout("data.dat");

    std::ofstream fcout("collision-totals.dat");
    fout << "set xlabel 'x'\nset ylabel 'y'\nset zlabel 'z'\n";
    std::string tree_plotopt = "w lp ps 1 t ''" ;
    std::string part_plotopt = "w p pt 6 ps 2 lt 2 t 'particles'";
    std::ostringstream ostr;


    fcout << "#  Collisions Totals\n"
             "#  Sim. Time(s)        inst. coll./s"
             "          avg. coll./s"
             "          avg. coll. selection rate (per atom)"
             "          Num. Super Particles"
             "          Cycle CPU Time(s)\n";

    std::ofstream fenergy("tot-energy.dat");
    std::ofstream frXp("tot-rXp.dat");


    std::ofstream vbinout[3];
    vbinout[X].open("vbinX.dat");
    vbinout[Y].open("vbinY.dat");
    vbinout[Z].open("vbinZ.dat");

    double max_dx[3] = {0.0, 0.0, 0.0};

    /* now let the dsmc have some fun with the particles. */
    try {
        signal(SIGINT, quitRequested);
        double seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);
        while ( !quit_requested && dsmc.getTotalTime() <= t_max) {
            dsmc.addParticles(atomsrc.createParticles(dt));

            struct tms ti, tf;
            times(&ti);
            dsmc.evolve(dt, NULL);
            times(&tf);

            if (dsmc.getTotalTime() >= sample_wait_time) spit_out_data = 1;

            {
                double cycle_cpu_time = 
                     ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
                       * seconds_per_clock_tick
                     );

                /* plot collision totals every time-step. */
                fcout << dsmc.getTotalTime()       << '\t'
                      << (dsmc.getCycleCollisions()/(dt*dsmc.numberParticles())) << '\t'
                      << (dsmc.getTotalCollisions()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                      /* the following is supposed to be collision selection rate
                       * per atom.  That is why the factor of 2 precedes the line.
                       */
                      << (2*dsmc.getTotalCollisionSelections()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                      << dsmc.numberParticles() << '\t'
                      << cycle_cpu_time << std::endl;
            }

            if ((int(dsmc.getTotalTime() / dt) - 1) % sample_interval  == 0) {

                for (int i = 0; i < dsmc.numberParticles(); i++) {
                    Vector<double,3> & v = V3C(dsmc.getParticles()[i]->vel);
                    vbin[X].bin(v[X]);
                    vbin[Y].bin(v[Y]);
                    vbin[Z].bin(v[Z]);
                }


            }

            /* **** TAKE CARE OF PLOTTING ALL PARTICLES. */
            if (dsmc.getTotalTime() >= (t_max-dt) || (int(dsmc.getTotalTime() / dt) - 1) % write_interval  == 0) {
                std::string tot_time = to_string(dsmc.getTotalTime());
#if 0
                /* note that this octree is not necessarily representative of the
                 * current particles.  This is because there has been a call to
                 * dsmc.evolve since the Octree was built.
                 */
                dsmc.plotParticlesAndTree(fout,
                                          tot_time + " s",
                                          tree_plotopt,
                                          part_plotopt)
    	            << "pause -1\n" << std::flush;
#endif

                double energy = 0.0;
                Vector<double,3> rXp(0.0);
                for (int i = 0; i < dsmc.numberParticles(); i++) {
                    Vector<double,3> & r = *((Vector<double,3>*)dsmc.getParticles()[i]->pos);
                    Vector<double,3> & v = *((Vector<double,3>*)dsmc.getParticles()[i]->vel);
                    energy += (v*v);
                    rXp += cross(r,v);
                }
                energy *= 0.5 * particle_info[particle_type].mass;
                rXp *= particle_info[particle_type].mass;

                fenergy << tot_time << "\t" << energy << std::endl;
                frXp << tot_time << "\t" << rXp << std::endl;

                if (dsmc.getInterpolator()) {
                    /* ok, let's spit out the data tree. */
#if 0
                    dsmc.getInterpolator()->
                         plotParticlesAndTree(fdata,
                                              "DSMC Data Tree:  " + tot_time + " s",
                                              tree_plotopt,
                                              part_plotopt)
                     << "pause -1\n" << std::flush;
#endif

#if 0
                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    dsmc.getInterpolator()->printParticles(fpout) << "\n\n";
#endif


                    /* We're going to normalize the bin before we write it
                     * out. */
                    int vbinsum[3] = {0,0,0};
                    for (unsigned int i = 0; i < vbin[X].nBins(); i++) {
                        vbinsum[X] += vbin[X].bins[i];
                        vbinsum[Y] += vbin[Y].bins[i];
                        vbinsum[Z] += vbin[Z].bins[i];
                    }

                    /* the factor of 1e7 is just to make sure that we have a
                     * dynamic range of 0:1e4 after converting back to
                     * integer. (Remember that there are 1000 bins in the
                     * histogram.) */
                    for (unsigned int i = 0; i < vbin[X].nBins(); i++) {
                        tmpvbin[X].bins[i] = (int)rint( 1e7 * ((double)vbin[X].bins[i]) / ((double)vbinsum[X]) );
                        tmpvbin[Y].bins[i] = (int)rint( 1e7 * ((double)vbin[Y].bins[i]) / ((double)vbinsum[Y]) );
                        tmpvbin[Z].bins[i] = (int)rint( 1e7 * ((double)vbin[Z].bins[i]) / ((double)vbinsum[Z]) );
                    }

                    /* write out new shmooshed together spline data. */
                    vbinout[X] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    vbinout[Y] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    vbinout[Z] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    tmpvbin[X].print(vbinout[X],tot_time + '\t') << "\n\n" << std::flush;
                    tmpvbin[Y].print(vbinout[Y],tot_time + '\t') << "\n\n" << std::flush;
                    tmpvbin[Z].print(vbinout[Z],tot_time + '\t') << "\n\n" << std::flush;

                    if ( !spit_out_data ) {
                        /* we'll just keep clearing the bins until we need to
                         * really start time-averaging.
                         */
                        vbin[X].clearBins();
                        vbin[Y].clearBins();
                        vbin[Z].clearBins();
                    }

                    /* now let's zero the collisions-per-octree-node counters
                     * so that we only get number collisions per sample time.
                     */
                    DSMCData<> ** d = dsmc.getInterpolator()->getParticles();
                    for (int i = dsmc.getInterpolator()->numberParticles() - 1; i >= 0; i--) {
                        d[i]->data.name.collisions = 0;
                    }
                } else if (dsmc.numberParticles() > 0 && spit_out_data) {
                    /* we have to make our own interpolator since DSMC didn't need
                     * to construct it. */
                    octree::OctreeInterpolator<DSMCData<>,Particle> ointerp;
                    DSMCData<>::InitInfo init = {dsmc.getParticles()};

                    ointerp.createDataOctree(dsmc.getOctree(), init);
                    ointerp.setRadiiOfInfluence(dsmc.data_radius_of_influence);

                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    ointerp.printParticles(fpout) << "\n\n";
                }

            }/* if */

            for (int i = X; i <= Z; i++)
                max_dx[i] = std::max(max_dx[i], dsmc.max_particle_transit(i));
        }/* while */
    } catch (std::exception & e) {
        std::cout << "exception:  " << e.what() << std::endl;
    }

    fout.close();
    fpout.close();
    fdata.close();
    fcout.close();

    fenergy.close();
    frXp.close();

    vbinout[X].close();
    vbinout[Y].close();
    vbinout[Z].close();

    std::cout << "max Particle Transit:  ("
              << max_dx[X] << ','
              << max_dx[Y] << ','
              << max_dx[Z] << ')'
              << std::endl;

    return 0;
}

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore) {
    /* we're going to leave this a 1d problem to compare to Birds code. */
    rkF[X] = p[VX];
    rkF[Y] = p[VY];
    rkF[Z] = p[VZ];
#  if defined(DIMENSIONS) && (DIMENSIONS < 3)
    rkF[Y] = 0.0;
#    if  (DIMENSIONS < 2)
    rkF[Z] = 0.0;
#    endif
#  endif

    /* no forces. */
    rkF[VX] = 0.0;
    rkF[VY] = 0.0;
    rkF[VZ] = 0.0;
}

