
#include <dsmc/ParticleNode.h>
#include <dsmc/OctreeContainer.h>
#include <dsmc/ParticleSrc.h>
#include <dsmc/DSMC.h>
#include <dsmc/debug_util.h>
#include <olson-tools/physical.h>
#include <olson-tools/logger.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

using namespace physical::constants;
using namespace physical::units;

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore);

int mainprog(int argc, char *argv[]);

int main(int argc, char *argv[]) {
    TRY {
        return mainprog(argc,argv);
    } CATCH(std::exception &e,
        std::cout << getpid() << ":  exception caught:  " << e.what() << std::endl;
    ) CATCH(...,
        std::cout << getpid() << ":  non-std:: exception caught." << std::endl;
    )

#ifndef _OPENMP
    return EXIT_FAILURE;
#endif
}

int mainprog(int argc, char *argv[]) { 
    setLogProgramName("mpi4dsmc");
    /* add some particles to the system. */
    int8_t particle_type = getParticleTypeIndex("87Rb");
    particle_info[particle_type].size = 1e10;
    ParticleSrc<> atomsrc;
    atomsrc.seed.type              = particle_type;
    atomsrc.seed.dt                =  10*us;
    atomsrc.pos_distrib.sigma_x[X] = 10. * mm;
    atomsrc.pos_distrib.sigma_x[Y] = 10. * mm;
    atomsrc.pos_distrib.sigma_x[Z] = 10. * mm;

    atomsrc.origin[X] = 0.0;
    atomsrc.origin[Y] = 0.0;
    atomsrc.origin[Z] = 0.0;

    atomsrc.temperature[X] = 100. * uK;
    atomsrc.temperature[Y] = 100. * uK;
    atomsrc.temperature[Z] = 100. * uK;

    atomsrc.velocity[X] = 1. * m/s;
    atomsrc.velocity[Y] = 0.;
    atomsrc.velocity[Z] = 0.;

    const double dt = 0.001;
    atomsrc.flux = 1e13 / dt;
    atomsrc.initialize();


    /* set up the DSMC algorithm. */
    ParticleNode<Particle,DSMC< RK2Integrator > > node;
    // node.amin_N_per_proc = 1000.0; /* defaults to 1000.0 */

    Vector<double,3> v1 = V3(-1.0,0.0,0.0),
                     v2 = V3(30.0*cm, 0.0, 0.0);
    octree::VacuumBoundaryObject<Particle> vwall(v1,v2);
    node.container.addBoundaryObject(&vwall);
    node.container.integrator.derivs = flatderivs;
    node.container.allow_collisions = true;
    node.container.data_radius_of_influence = 4;

    /* here goes nothing... */
    node.init(argc, argv);

    /* the init(...) command creates our own subdirectory to run from.  Now we
     * can open up our own versions of output files. */

    LogFileP = fopen("logger.out", "a");
    if (!LogFileP) {
        log_severe("cannot open log file");
    }

    std::ofstream nodeinfo("node_info");
    node.printInfoHeader(nodeinfo << "# ") << std::endl;


    for (double t = 0.; t < 0.401; t+= dt) {
        if (node.mpi_info.arank >= 0) {
            node.addParticles(atomsrc.createParticles(dt));
        }
        node.evolve(dt);

        node.printInfo(nodeinfo) <<  std::endl;
    }/* for */
    return 0; 
} 

void flatderivs(const double p[VZ+1], const double * time, const double * dt, double rkF[VZ+1], void * ignore) {
    rkF[X] = p[VX];
    rkF[Y] = p[VY];
    rkF[Z] = p[VZ];

    /* no forces. */
    rkF[VX] = 0.0;
    rkF[VY] = 0.0;
    rkF[VZ] = 0.0;
}

