// -*- c++ -*-
// $Id: DSMC.C 474 2008-05-06 13:20:03Z olsonse $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: DSMC.C,v $
 * Revision 1.3  2005/06/07 10:00:04  olsonse
 * Many changes over the course of time
 *
 * Revision 1.2  2005/01/19 04:29:56  olsonse
 * Tree upgrade now only includes one call to OctreeInterpolator::interpolate.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:26  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#include "DSMC.h"
#include <olson-tools/physical.h>
#include <olson-tools/m_eps.h>
#include <vector>

int dsmc_c_rcsid_dummy = olson_tools::add_rcsid("$Id: DSMC.C 474 2008-05-06 13:20:03Z olsonse $");

const double ONE_10EPS = 1.0 + (10. * M_EPS);

DSMCTemplate
DSMCClass::DSMC():
              super(),
              ointerp(NULL),
              total_collision_events(0),
              cycle_collision_events(0),
              number_collision_pair_selections(0),
              number_evolutions(0),
              total_cycles(0),
              total_time(0.0),
              absolute_time(0.0),
              recent_max_collision_rate(0.0),
              integrator(),
              allow_collisions(true),
              data_radius_of_influence(10),
              leafModel(),
              dt_min(0.0) {
    max_dx.zero();
};

DSMCTemplate
DSMCClass::~DSMC() {
    if (ointerp) delete ointerp;
}

#define SqrVelocity(ma)	(SQR((ma)->vel[X]) + SQR((ma)->vel[Y]) + SQR((ma)->vel[Z]))

#define RelVelCmpSQR(a1,a2,j)	SQR((a1)->vel[j] - (a2)->vel[j])

#define RelativeSpeed(a1,a2)	sqrt(RelVelCmpSQR(a1,a2,X) + RelVelCmpSQR(a1,a2,Y) + RelVelCmpSQR(a1,a2,Z))

const int DT_UPGRADE_LEN = 4;
const int DT_UPGRADE[DT_UPGRADE_LEN] = {
    DSMCData<>::NUMBER_COLLISIONS,
    DSMCData<>::NUMBER_COLLISION_PAIRS_REMAINDER,
    DSMCData<>::MAX_SIGMA_RELSPEED,
    DSMCData<>::NUMBER_DENSITY
#if 0
    ,
    DSMCData<>::AVG_V_X,
    DSMCData<>::AVG_V_Y,
    DSMCData<>::AVG_V_Z,
    DSMCData<>::AVG_V2_X,
    DSMCData<>::AVG_V2_Y,
    DSMCData<>::AVG_V2_Z
#endif
};

const std::vector<int> DTV_UPGRADE(DT_UPGRADE, DT_UPGRADE+DT_UPGRADE_LEN);

DSMCTemplate
inline void DSMCClass::upgradeDataTree() {
    /* create data tree to represent the current particle configuration. */
    struct DSMCData<_Particle>::InitInfo_s init = {super::getParticles()};
    octree::OctreeInterpolator<DSMCData<_Particle>,_Particle> * newointerp = new octree::OctreeInterpolator<DSMCData<_Particle>,_Particle>;
    newointerp->createDataOctree(super::getOctree(), init);
    newointerp->setRadiiOfInfluence(this->data_radius_of_influence);

    if (!ointerp) {
        /* initialize the first data tree.
         * TODO:  make sure that everything is properly initialized.
         * OctreeInterpolator::createDataOctree does a fair portion of this in
         * DSMCData<_Particle>::DSMCData(...).
         */
    } else {
        /* properly upgrade the data tree to the current particle
         * configuration. */
        DSMCData<_Particle> ** data = newointerp->getParticles();

        for (int i = newointerp->numberParticles()-1; i >= 0; i--) {
            DSMCData<_Particle> & d = *data[i];
            std::vector<double> interpolant = ointerp->interpolate(d.pos, DTV_UPGRADE);

            /* we only use the interpolated values if we were inside the
             * kernel's reach (@see OctreeInterpolator::*).
             */
            if (interpolant.size() > 0) {
                d.data.name.collisions = interpolant[0];
                d.data.name.number_collision_pairs_remainder = interpolant[1];

                /*FIXME:  come up with better/more correct weights for the
                 * following two running averages. */

                /* for max_sigma_relspeed, we'll do a weighted average of the
                 * interpolated value and something derived from the temperature. */
                d.data.name.max_sigma_relspeed *= 0.05;
                d.data.name.max_sigma_relspeed += 0.95*interpolant[2];
//              d.data.name.max_sigma_relspeed = interpolant[2];

                /* for numberDensity, we'll do a weighted average of the
                 * interpolated value and current density measurement.
                 */
                d.data.name.numberDensity *= 0.15;
                d.data.name.numberDensity += 0.85*interpolant[3];

#if 0
                /* for <v> and <v^2> we do a weighted average. */
                d.data.name.avgv[X] *= 0.95;
                d.data.name.avgv[Y] *= 0.95;
                d.data.name.avgv[Z] *= 0.95;
                d.data.name.avgv[X] += 0.05*interpolant[4];
                d.data.name.avgv[Y] += 0.05*interpolant[5];
                d.data.name.avgv[Z] += 0.05*interpolant[6];

                d.data.name.avgv2[X] *= 0.95;
                d.data.name.avgv2[Y] *= 0.95;
                d.data.name.avgv2[Z] *= 0.95;
                d.data.name.avgv2[X] += 0.05*interpolant[7];
                d.data.name.avgv2[Y] += 0.05*interpolant[8];
                d.data.name.avgv2[Z] += 0.15*interpolant[9];
#endif
            }/* if */
        }/* for */

        /* delete the old data tree. */
        delete ointerp;
    }

    /* now assign the new tree to the ointerp pointer. */
    ointerp = newointerp;
}

DSMCTemplate
inline void DSMCClass::evolve(const double & dt, void * forceargs) {
    /* make sure the removal queue was emptied first. */
    super::flushRemovalQueue();

    /* update the total number of cycles and time */
    total_cycles++;
    cycle_collision_events = 0;

    /* the following do loop is an attempt at adjusting the timestep so-as to
     * avoid doing too large of timesteps for the densities at hand. */
    double dt_remaining = dt;
    static double dt_current = dt;
    double dt_real = 0.0;
    /* we'll throw any extra time that is up to 10*M_EPS more than needed. */
    max_stuff.zero();
    do {
        /* no use trying to continue without any particles */
        if (!rebuildOctree()) {
            total_time += dt_remaining;
            absolute_time += dt_remaining;
            break; /* we're finished. */
        }

        /* we want to try and force all timesteps to be roughly (1/10) of the
         * collision time.  I don't yet have a good way to have multiple
         * timesteps as a function of position.  What would we do for movement
         * between MPI nodes?  As it stands, there will grow a discrepancy
         * between MPI ParticleNode instances if dt is always much larger
         * than (1/10) of collision time.  Not sure what to do about this
         * right now.  */

        /* some more hack code to slow down the adaptation. */
        double ddt = 0.1/(1e-100 + recent_max_collision_rate);
        double dt_err = std::min(100.0,std::max(.01,dt_current/ddt));
        dt_current = std::min(dt, std::max(dt_min, dt_current * fast_pow(dt_err, -.5)));
        dt_real = dt_remaining / (1.0 + (double)std::max(0,int(dt_remaining/dt_current - 1e-3)));

#if defined (MONITOR_COLLISION_RATE)
        for (unsigned int ii = 0; ii < monitors.size(); ii++) {
            monitors[ii].startMeasure(super::getParticles()[0]->type);
        }/* for */
#endif

        /* take care of all necessary collisions */
        if (allow_collisions || super::boundaryObjects().size() > 0) {
            /* we must first do whatever necessary to obtain the accurate data
             * tree. */
            upgradeDataTree();
        }

        /* reset the measurement of the maximum transit in the DSMC container */
        max_dx.zero();

        /* reset the maximum collision rate measurement. */
        recent_max_collision_rate = 0.0;

        /* and now recurse down to:
         * - do collisions
         * - move particles
         * - interact with boundaries. 
         */
        evolve(dt_real, forceargs, *super::getOctree());

        /* now let's clean up a little. */
        super::flushRemovalQueue();

#if defined (MONITOR_COLLISION_RATE)
        for (unsigned int ii = 0; ii < monitors.size(); ii++) {
            monitors[ii].finishMeasure();
        }/* for */
#endif

        /* now update the total time */
        total_time += dt_real;
        absolute_time += dt_real;
        dt_remaining -= dt_real;
    } while (dt_remaining * ONE_10EPS > dt_real) ;
}

/* for the sake of being easy to read, we'll do this as a recursive function
 * for now.  In the future, we may want to flatten this into one non-recursive
 * function.
 */
DSMCTemplate
template<class _OctreeType>
void DSMCClass::evolve(const double & dt, void * forceargs, _OctreeType & otree ) {
    if (otree.thisHasChildren) {

        /* not a leaf node, so continue recursing to children */
        for (int i = 0; i < 8; i++) {
            if (otree.child[i]) evolve(dt, forceargs, *otree.child[i]);
        }

    } else {
        DSMCData<_Particle> * dptr = NULL;
        std::vector< octree::BoundaryObject<_Particle> *> bo_list;

        if (allow_collisions || super::boundaryObjects().size() > 0) {
            /* FIXME:  If we indeed continue to upgrade the data tree to the
             * new octree locations EVERY stinking time-step, there is no need
             * to do "findNearestLeaf".  In this case, we should be able to
             * simply do a lookup.  There should be a generic way of storing a
             * pointer to the corresponding data leaf-node in each particle
             * leaf node.  Perhaps even join the two trees and extend the
             * CMVolMeasure class used for DSMC to measure all the other
             * relevant stuff and store it in the particle leaf-node in the
             * "extra" field.  OctreeInterpolator would have to be changed to
             * know where to find this information.  I would rather the limit
             * the interpolators ability to work with a data-only tree, so I
             * would need a <humor me>generic</humor me> way of instructing
             * the interpolator how to find its data.  Actually, duh!, this
             * should be easy:  for a data only tree, we "create" particles
             * for each data point, do an octree on that data point and then
             * copy in the appropriate data.  Taking this approach, I would no
             * longer use the term "data particle", as data objects would no
             * longer be required to have any position information, just
             * whatever data they needed.  The only caveat is that for using a
             * data-only tree, the position for each bit of data will need to
             * be specified.  This is reasonable.  
             * It follows also from this approach that it no longer makes
             * sense to have the interpolator be a seperate container, but
             * rather a tool to use with special octrees--those with the
             * appropriate "extra" fields.  The "extra" fields will need to
             * inherit from the InterpolatorFields class, or at least include
             * a "double maxRadiusOfInfluence" field--easier to just inherit. 
             *
             * Another thought:  instead of upgrading the data elements
             * separately after the tree is built, just allow each call to
             * DSMCCMVolMeasure to add in the historical contributions via
             * ointerp->interpolate(...).  This way, we don't need to do
             * yet another recursion through the tree.  The only portion that
             * will need to be rethought is the setRadiiOfInfluence(...)
             * routine that currently walks up the tree several times starting
             * from each leafnode.
             *
             * Yet another though:  I am not really convinced that we need to
             * smooth out numberDensity in order to get the DSMC correct.  We
             * already do a reasonable job of dampening oscillations of N per
             * node because of the octree.  In addition, as I've looked at the
             * computed value of numberDensity runtime in a debugger, I
             * believe I've seen that the instantaneous values were either
             * the same or better matched to expected values (especially
             * during stream start-up conditions).
             * So, if it doesn't really seem to help to smooth out
             * numberDensity, we ought to just use the instantaneous value.
             *
             * Continuing this thought:
             * The only other two things that require the tree are
             * max_sigma_relspeed and number_collision_pairs_remainder.
             * 1.  Why in do we really need max_sigma_relspeed?  Can't we just
             * use the temperature which we already measure instantaneously?
             * Just do something like assume that
             * max(vrel) = 4*sigma_v = * 4*sqrt(3 K_B T / m)
             * 2.  I'm not sure how to eliminate
             * number_collision_pairs_remainder yet, except:  do we really
             * need it?  Is there a way to rescale that last bit of
             * probability to remove the need to keep this remainder around?
             * Perhaps I can do something similar to what I do in the
             * laser-cooling code:
             * Allow the remainder to be probabilistically up-converted to a
             * whole integer or down-converted to zero:  pick a random number
             * between one and zero.  If the remainder is less than this
             * random number, up-convert, otherwise, down-convert.  
             *
             * Dissimilar to what Bird has in his book, we will almost never have too few
             * particles per node to allow a collision in a particular node
             * because we disallow the situation.
             *
             * If (1) and (2) are valid, then we can get rid of the data-tree
             * and interpolation altogether and this code would improve in
             * time-accuracy.  
             */
            dptr = ointerp->getParticles()[ointerp->getOctree()->findNearestLeaf(otree).low];
        }

        if (allow_collisions) {
            /* Yeah! DSMC actually needed for something. */
            doCollisions(dt, otree, dptr);
        }

        if (super::boundaryObjects().size() > 0) {
            Vector<double,3> max_dx_;
            const Vector<double,3> & vav = V3C(dptr->data.name.avgv);
            const Vector<double,3> & v2av = V3C(dptr->data.name.avgv2);

            /* We use an estimated maximum transit size to minimize the
             * number of boundaries to interact with.
             * This estimated value is found by:
             * dt( |<v>| + sqrt( sum(<v.^2>) - <v>*<v> ) ) + dxHook
             * 
             * The three sources of this estimate are reasoned as follows:
             * 1.  dt * |<v>| gives maximum transit based on average velocity.
             * 2.  sqrt(...) gives us the std-dev of the velocity.  We'll
             * assume 20 \sigma_{v} is max velocity such that we have <<1%
             * miss rate for objects. 
             * 3.  allow the user to hook in additional translation lengths due to
             * things like accelerations.
             *
             * Note that although the OctreeContainer code allows you to scale
             * the dimensions asymmetrically, I choose to scale symmetrically
             * because reflections from nearby objects can cause particles to
             * miss other objects during the same timestep.  
             * Other note:  the 1.5 factor helps us avoid precision problems.
             */

            max_dx_ = 1.5*dt*( vav.abs() + 20.0 * sqrt(ONE_10EPS*sum(v2av) - (vav*vav)) )
                    + dxHook::dxHook(dt,forceargs,otree,*this);
            
            minimizeBoundaryObjects(bo_list, otree, max_dx_.val);
        }

        /* and now we take care of moving particles in this leaf-node as well
         * as interacting with boundaries. */

        /* take care of all simple movements;
         * let's have them moved to their new coordinates.
         * */
        for (int i = otree.low; i <= otree.high; i++) {
            _Particle & p = *super::getParticles()[i];
            _Particle p_old(p); /* a record of the particle before movement. */
            integrator.integrate(VNCAST(double,6,p.pos),
                                 absolute_time,
                                 dt,
                                 p.dt,/*only used by adaptive routines. */
                                 forceargs);
            number_evolutions++;

            Vector<double,3> dx(V3C(p.pos));
            dx -= V3C(p_old.pos);
            dx.save_fabs();

            /* update the maximum transit result. */
            if (max_dx[X] < dx[X]) max_dx[X] = dx[X];
            if (max_dx[Y] < dx[Y]) max_dx[Y] = dx[Y];
            if (max_dx[Z] < dx[Z]) max_dx[Z] = dx[Z];

            /* have the container apply boundary conditions. */
            super::interactParticleWithBoundaryObjects(p_old, i, bo_list);

#if defined (MONITOR_COLLISION_RATE)
            for (unsigned int ii = 0; ii < monitors.size(); ii++) {
                if (monitors[ii].boundedBy(p)) {
                    monitors[ii].addMeasure(p);
                }
            }/* for */
#endif
        }/* for */
    }// else is a leaf node
}/* recursive evolve */

#ifdef DEBUG_COLL_RATE
std::ofstream debugcoll;
unsigned long long debug_num_collisions = 0;
unsigned long long debug_iter = 0;
#endif

DSMCTemplate
template <class _OctreeType>
inline void DSMCClass::doCollisions(const double & dt, _OctreeType & otree, DSMCData<_Particle> * dptr) {
    /* and the number of collision pairs to select is ...
     * ( from equation 11.3 of Bird )
     * */

    if (!dptr) {
        dptr = ointerp->getParticles()[ointerp->getOctree()->findNearestLeaf(otree).low];
    }

    DSMCData<_Particle> & d = *dptr;

    /* FIXME:  is the generic solution of this just to allow for
     * multi-species collisions?
     * NoTe:  d.data.name.numberDensity is a smoothed quantity (in space and
     * time).
     */
    double number_coll_selections =
        (     0.5
            * double(otree.numberParticles()) /* N */
            * d.data.name.numberDensity /* * otree.getVolume() */ /* \bar{N} */
            * particle_info[d.particle_type].size
            * d.data.name.max_sigma_relspeed
            * dt
       /*   / otree.getVolume() */
        )
        + d.data.name.number_collision_pairs_remainder;

#ifdef DEBUG_COLL_RATE
#if 0
    number_coll_selections *=5.e-16 /  d.data.name.max_sigma_relspeed;
    d.data.name.max_sigma_relspeed = 5.e-16;
#endif
    //number_coll_selections *=1.08;
    debug_num_collisions = 0;
    {
        static bool init = false;
        if (!init)  {
            debugcoll << "# iter\tnsel\tN\tn\tF\tmax(sv_r)\tdt\trem(..)\tV\tnColl\n";
            init = true;
        }
        debugcoll << debug_iter << '\t'
              << number_coll_selections << '\t'
              << double(otree.numberParticles()) /* N */ << '\t'
              << d.data.name.numberDensity  << '\t'
              << double(particle_info[d.particle_type].size) << '\t'
              << d.data.name.max_sigma_relspeed << '\t'
              << dt << '\t'
              << d.data.name.number_collision_pairs_remainder << '\t'
              << otree.getVolume() <<  '\t';
    }
    debug_iter++;
#endif

    d.data.name.number_collision_pairs_remainder = number_coll_selections
                                                 - int(number_coll_selections);

    if (number_coll_selections >= 1) {
        if (otree.numberParticles() < 2) {
            /* hmm, not even enough atoms in the cell to
             * do one collisions!
             * */
            d.data.name.number_collision_pairs_remainder = number_coll_selections;
        } else {
            /* I'm going to try something different than
             * bird here:  I also allow for a downgrade of max_sigma_relspeed.
             * I'm thinking of doing this so that each
             * cell can have its own (lower) temperature,
             * because we do hope for cooling through the
             * guide. */
            double max_sigma_relspeed = 0.0;
            unsigned long long num_collisions = 0;

            number_collision_pair_selections += int(number_coll_selections);
            for ( int j = int(number_coll_selections); j > 0; j-- ) {

                /* select two probable collision partners */
                int p1=-1,p2=-1;
                selectCollisionPartners(otree, p1, p2);

                    double sigma_relspeed;
                {
                    register double v_rel = RelativeSpeed(super::getParticles()[p1], super::getParticles()[p2]);
                    sigma_relspeed
                      = v_rel

                            /* the collision cross-section is based on
                             * eqn (4.63) for VHS model. */
                      * particle_info[d.particle_type].cross_section
                      * fast_pow(
                          ( 2.0 * physical::constant::K_B * particle_info[d.particle_type].T_ref
                            / ( 0.5 * particle_info[d.particle_type].mass /* reduced mass */
                                * v_rel*v_rel
                              )
                          ), (particle_info[d.particle_type].visc_T_law - 0.5))
                      * particle_info[d.particle_type].gamma_visc_inv;
                }

                /* if necessary, the maximum product in max_sigma_relspeed is upgraded */
                if (sigma_relspeed > max_sigma_relspeed) {
                    max_sigma_relspeed = sigma_relspeed;
                }

                if (MTRNGrand() < (sigma_relspeed/d.data.name.max_sigma_relspeed) ) {
                    num_collisions++;

                    /* mix the momenta of the particles. */
                    CollisionModel::collide(*(super::getParticles()[p1]),*(super::getParticles()[p2]));

#if defined (MONITOR_COLLISION_RATE)
                    for (unsigned int i = 0; i < monitors.size(); i++) {
                        if (monitors[i].boundedBy(*super::getParticles()[p1]) &&
                            monitors[i].boundedBy(*super::getParticles()[p2])   ) {
                            monitors[i].data.d.data.name.collisions+=2;
                        }/* if */
                    }/* for */
#endif
                }/* if */
            }/* for */

            /* book keeping: */
            d.data.name.collisions+=2*num_collisions; /* increment by two because two particles collided. */
            total_collision_events+=num_collisions;
            cycle_collision_events+=num_collisions;

#if 0
            recent_max_collision_rate =
                std::max(recent_max_collision_rate,
                         (2.0*num_collisions) / ( dt * otree.numberParticles())
                );
#else
            {
                double my_coll_rate =
                    (2.0*num_collisions) / (dt * otree.numberParticles());
                recent_max_collision_rate =
                    std::max(recent_max_collision_rate, my_coll_rate);

                if (my_coll_rate > max_stuff.collision_rate) {
                    max_stuff.collision_rate = my_coll_rate;
                    max_stuff.pos = otree.getRcm();
                    max_stuff.N = otree.numberParticles();
                    max_stuff.volume = otree.getVolume();
                    max_stuff.dt = dt;
                    max_stuff.number_coll_selections = number_coll_selections;
#ifdef DEBUG_COLL_RATE
                    max_stuff.debug_iter = debug_iter;
#endif
                }
            }
#endif

            d.data.name.max_sigma_relspeed *= 0.5;
            d.data.name.max_sigma_relspeed += 0.5*max_sigma_relspeed;

#ifdef DEBUG_COLL_RATE
            debug_num_collisions = num_collisions;
#endif
        }/* else */
    }//if int(number selections) > 0

#ifdef DEBUG_COLL_RATE
    debugcoll << debug_num_collisions << std::endl;
#endif
}/* doCollisions */



DSMCTemplate
inline int DSMCClass::splitParticles(
                        const double & old_size,
                        const double & new_size,
                        const Vector<double,3> & density) {
    _Particle * new_particles = NULL;
    _Particle ** nPtr = &new_particles;

    for (int j = super::numberParticles()-1; j >= 0; j--) {
        *nPtr = split(old_size, new_size, density, *super::getParticles()[j]);

        /* nPtr needs to point at the last valid Particle pointer. */
        while (*nPtr) nPtr = &((*nPtr)->next);
    }

    /* FIXME:  now, we need to ensure that all particles are indeed still
     * within the boundaries. (do we?)
     */

    return super::addParticles(new_particles);
}

