// -*- c++ -*-
// $Id: DSMC.h 583 2008-07-21 17:16:04Z olsonse $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: DSMC.h,v $
 * Revision 1.2  2005/06/07 10:00:05  olsonse
 * Many changes over the course of time
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 */

#ifndef DSMCCONTAINTER_H
#define DSMCCONTAINTER_H

#include "Particle.h"
#include "DSMCData.h"
#include "OctreeContainer.h"
#include "OctreeInterpolator.h"
#include "collide.h"
#include <olson-tools/random/random.h>
#include <olson-tools/RKIntegrator.h>
#include <olson-tools/Vector.h>
#include <olson-tools/version.h>

#if defined (MONITOR_COLLISION_RATE)
#  include "OctreeMonitor.h"
#  include "DSMCMonitor.h"
#  include <vector>
#endif

int dsmc_h_rcsid_dummy = olson_tools::add_rcsid("$Id: DSMC.h 583 2008-07-21 17:16:04Z olsonse $");

/** Do nothing hook for upgrading the maximum translation expected for a
 * certain temperature gas with a certain average velocity.
 */
class NULLDXHook {
  public:
    template <class OTree, class dsmc>
    static inline double dxHook(const double & dt,
                                void * forceargs,
                                OTree & o,
                                dsmc & d) {
        return 0.0;
    }
};

/** Hook for calculating the maximum translation due to an acceleration.
 * This is then added to the maximum translation expected for a certain
 * temperature gas with a certain average velocity. 
 * @param ASrc
 *     A source for an acceleration.
 */
template <class ASrc>
class AccelDXHook {
  public:
    template <class OTree, class dsmc>
    static inline double dxHook(const double & dt,
                                void * asrc,
                                OTree & o,
                                dsmc & d) {
        Vector<double,3> a;
        ((ASrc*)asrc)->accel(a, o.getRcm());
        return a.abs() * 0.5 * SQR(dt);
    }
};

/** The leaf model to determine the particle distribution per octree leaf.
 * This particular model allows the density to determine the size of the leaf
 * nodes.  The target is to make leaf nodes be 
 * \f$ \lesssim 0.5 \lambda_{mfp} \f$
 * on an edge.  With this heuristic, the burden of proof of computational
 * validity rests strictly on the number of particles in the system.  If this
 * limit is routinely met, then the number of particles should be increased.    
 *
 * Assuming that particles are randomly and evenly distributed in any leaf
 * node, we can expect the following mean-collision-separation:mean-free-path
 * ratio:
 *
 * the average value of the difference between two random numbers less that
 * 0.5 will be (1/6) with an rms value of 0.204.
 *
 * simple matlab code to check this:
 * <code>
 * N=100000;
 * s=.5; v=0; v2=0;
 * for ii=1:N;
 *     vi=(s*rand - s*rand);
 *     v+=abs(vi);
 *     v2+=vi.^2.0;
 * end;
 * v2=v2/N; v=v/N;
 * [ v, sqrt(v2)]
 * </code>
 *
 * @see Alexander, Garcia, Alder, ``Cell size dependence of transport
 * coefficients in stochastic particle algorithms,'' Phys. Fluids 10, p. 1540
 * (1998).
 */
class DSMCLeafModel {
  public:
    double min_N_leaf;
    double max_N_leaf;
    uint8_t particle_type;

    /** The maximum desired scale size of leaf nodes in units of the local
     * mean free path. */
    double max_scale_size;

    DSMCLeafModel() :
        min_N_leaf(4.0),
        max_N_leaf(100.),
        particle_type(0),
        max_scale_size(0.5) { }

    template <class _Octree>
    double maxParticlesPerLeaf(const _Octree & o) const {
        double number_density = o.numberParticles() / o.getVolume();

        double N_leaf = number_density
                      * fast_pow(
                            max_scale_size
                          / ( M_SQRT2 * particle_info[particle_type].cross_section
                              * number_density * particle_info[particle_type].size
                            ),
                            o.getDimensions()
                        );

        N_leaf = std::min(max_N_leaf, std::max(min_N_leaf, N_leaf));
        return N_leaf;
    }
};

std::ostream & operator<< (std::ostream & out, const DSMCLeafModel & dlm) {
    out << dlm.min_N_leaf << '\t' << dlm.max_N_leaf << '\t';
    return out;
}


#define DSMCTemplate     template < class Integrator, class CMAndVolMeasurement, class dxHook, class _Particle, class CollisionModel >
#define DSMCTemplateArgs Integrator,CMAndVolMeasurement,dxHook,_Particle,CollisionModel
#define DSMCClass        DSMC<DSMCTemplateArgs>


/** An implementation of a DSMC container and algorithm.
 * NOTE:
 * I mixed the new/delete calls with realloc/free calls in this class.
 * I use realloc because it makes it more convenient to reallocate space for
 * the Particle list as it grows.
 *
 * @param Integrator
 *     The particle position integrator used to provide Hamiltonian evolution.
 *     Note that the particle integrator will be passed the absolute_time and
 *     not total_time.
 *
 * @param CMAndVolMeasurement
 *     The type of octree subdivision to pursue.
 *
 * @param dxHook
 *     A hook for the user to use when boundaryobject testing is done.  The
 *     temperature and stream velocity are used naturally to determine the
 *     maximum translation.  This hook allows the user to also include
 *     additional translation due to acceleration.  In strong forces, this is
 *     important because most of the translation of a particle within an
 *     octree node might be due to its acceleration during the time step.
 * 
 * @param _Particle
 *     Specify the type of particle to use with the DSMC algorithm.
 *
 * @param CollisionModel
 *     Specify the type of collision model to use.  This allows specifying a
 *     non-elastic type of collision.  
 */
template < class Integrator = RK5AdaptiveIntegrator<> ,
           class CMAndVolMeasurement = octree::CMAndVolNotShrink,
           class dxHook = NULLDXHook,
           class _Particle = Particle,
           class CollisionModel = ElasticCollisionModel>
class DSMC: public octree::OctreeContainer<_Particle,octree::nothing,CMAndVolMeasurement,DSMCLeafModel> {
  public:
    typedef octree::OctreeContainer<_Particle,octree::nothing,CMAndVolMeasurement,DSMCLeafModel> super;

    /** Default constructor. */
    DSMC();

    /** Destructor. */
    ~DSMC();

    /** Get the total number of collision events that have happened so far.
     * @return Number of collision events executed thus far.
     */
    inline const unsigned long long & getTotalCollisionEvents() const { return total_collision_events; }

    /** Get the total number of collisions that have happened so far.
     * @return Number of collisions executed thus far.
     */
    inline unsigned long long getTotalCollisions() const { return 2*total_collision_events; }

    /** Get the number of collisions that have happened in this last cycle.
     * @return Number of collisions executed in last cycle.
     */
    inline const unsigned long long & getCycleCollisionEvents() const { return cycle_collision_events; }

    /** Get the number of collisions that have happened in this last cycle.
     * @return Number of collisions executed in last cycle.
     */
    inline unsigned long long getCycleCollisions() const { return 2*cycle_collision_events; }

    /** Get the total number of collision pair selections thus far.
     * Note that getTotalCollisions() == 2*getTotalCollisionEvents() <= getTotalCollisionSelections().
     * @return Number of collision pair selections thus far.
     */
    inline const unsigned long long & getTotalCollisionSelections() const { return number_collision_pair_selections; }

    /** Get the total number of Hamiltonian evolutions thus far.
     * @return Number of total evolutions thus far.
     */
    inline const unsigned long long & getTotalEvolutions() const { return number_evolutions; }

    /** Get the number of cycles that have happened thus far.
     * The details of this function or cycle definition are a little muddy
     * right now.
     * @return Number of cycles thus far.
     */
    inline const unsigned long long & getTotalCycles() const { return total_cycles; }

    /** Get the total time since beginning the DSMC simulation.
     * @return Total run time.
     */
    inline const double & getTotalTime() const { return total_time; }

    /** Get the absolute time.  Unless changed by the user via
     * setAbsoluteTime(...), this function returns a value identical to
     * getTotalTime().
     *
     * @return Absolute time.
     *
     * @see setAbsoluteTime
     * @see getTotalTime
     */
    inline const double & getAbsoluteTime() const { return absolute_time; }

    /** Set the absolute time.
     *
     * @see getAbsoluteTime
     */
    inline void setAbsoluteTime(const double & t) { absolute_time = t; }

    /** Get a pointer to the interpolator.
     * @return Pointer to interpolator, or NULL if there is none.
     */
    inline const octree::OctreeInterpolator<DSMCData<_Particle>,_Particle> * getInterpolator() const { return ointerp; }

    /** Instruct the module to use Hamiltonian evolution on all of its atoms.
     * @param dt
     *     If adaptive steps are used, then this is only limited by some
     *     fraction of the collision time.
     *
     * @param forceargs
     *     Arguments to the derivs function [Defaults to NULL].
     *
     * @see derivs.
     * */
    inline void evolve(const double & dt, void * forceargs = NULL);

    /** Split all the particles into smaller pieces of a new size.
     * @param old_size
     *     The old size of the particles.
     *
     * @param new_size
     *     The new size of the particles.
     *
     * @param density
     *     Specify each of the linear densities to use when creating the new
     *     particles.  A linear density of DBL_MAX is necessary to eliminate
     *     a distribution in a specific dimension.
     *
     * @returns
     *     The number of new particles created.
     *
     * FIXME:  make sure that all new particles are contained within the
     * boundaries.  Remove those that are not contained(?).
     */
    inline int splitParticles(const double & old_size,
                              const double & new_size,
                              const Vector<double,3> & density = DBL_MAX );

  private:
    template<class _OctreeType>
    inline void selectCollisionPartners(const _OctreeType & o, int & p1, int & p2) {
	/* choose the first atom */
	p1 = o.low + int(MTRNGrand() * (o.numberParticles()-0.0001));
	while ( p1 == (p2 = o.low + int(MTRNGrand() * (o.numberParticles()-0.0001)) ));
    }

    template<class _OctreeType>
    void evolve(const double & dt, void * forceargs, _OctreeType & octree );

    template<class _OctreeType>
    inline void doCollisions(const double & dt, _OctreeType & octree, DSMCData<_Particle> * dptr = NULL );
    inline void upgradeDataTree();

  public:
    inline const double & max_particle_transit(const int & idx) const {
        return max_dx[idx];
    }

    /** Rebuild the octree in this DSMC container using the correct LeafModel. */
    inline bool rebuildOctree() {
        if (!super::numberParticles()) {
            return false;
        }

        /* for now, we'll just use the single type of particle for this leaf
         * model. */
        leafModel.particle_type = super::getParticles()[0]->type;
        super::freeOctree();
        super::buildOctree(leafModel);
        return true;
    }

  private:
    octree::OctreeInterpolator<DSMCData<_Particle>,_Particle> * ointerp;
    unsigned long long total_collision_events;
    unsigned long long cycle_collision_events;
    unsigned long long number_collision_pair_selections;
    unsigned long long number_evolutions;
    unsigned long long total_cycles;
    double             total_time;
    double             absolute_time;
    Vector<double,3>   max_dx;

    /** Recent maximum collision rate throughout the entire DSMC container.
     * This number is used to help reduce the intermediary timestep when
     * collisions are too high.
     */
    double recent_max_collision_rate;

  public:
#if 1
    class MaxStuff {
      public:
        MaxStuff() { zero(); }
        void zero() {
            collision_rate = 0.0;
            pos = 0.0;
            N = 0.0;
            volume = 0.0;
            dt = 0.0;
            number_coll_selections = 0.0;
        }
        double collision_rate;
        Vector<double,3> pos;
        double N;
        double volume;
        double dt;
        double number_coll_selections;
#ifdef DEBUG_COLL_RATE
        unsigned long long debug_iter;
#endif
    };
    MaxStuff max_stuff;

#endif
    /** Integrator class used to move the particles individually. */
    Integrator integrator;

    /** Set whether collisions are to be allowed or not. */
    bool		allow_collisions;

    /** Radius of influence to use when building the OctreeInterpolator data
     * tree.
     * This is in units of number of neighboring nodes to consider in the
     * interpolation.
     */
    int                 data_radius_of_influence;

    /** Specifies how the leafs are to be built.
     * @see OctreeContainer::buildOctree and Octree::buildOctree.
     */
    DSMCLeafModel leafModel;

    /** minimum step size for adaptive step (Defaults 0). */
    double dt_min;

#if defined (MONITOR_COLLISION_RATE)
    std::vector< octree::OctreeMonitor<DSMCMonitor<_Particle>,_Particle> > monitors;
#endif
};

#include "DSMC.C"

#endif // DSMCCONTAINTER_H

