// -*- c++ -*-
// $Id: DSMCMonitor.h,v 1.1.1.1 2005/01/08 04:57:25 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: DSMCMonitor.h,v $
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 */


/** \file
 * Particle type class definitions for monitoring DSMC and use with OctreeMonitor.
 * Copyright 2004 Spencer Olson.
 */


#ifndef DSMCMONITOR_H
#define DSMCMONITOR_H

#include <ostream>
#include <stdlib.h>
#include "common.h"
#include "DSMCData.h"
#include "Particle.h"
#include <olson-tools/physical.h>


/** A particle class to hold data values in an octree.
 * */
template <class _Particle = Particle>
class DSMCMonitor {
  public:
    inline DSMCMonitor(const double & max = 5*physical::unit::m/physical::unit::s,
                         const unsigned int & nb = 1000) :
        nbins(nb), max_bin_v_inverse(1/max) {
        v_bin = new int[nbins];
        memset(v_bin, 0, sizeof(int)*nbins);
    }
    inline DSMCMonitor(const DSMCMonitor & that) {
        *this = that;
    }

    inline const DSMCMonitor & operator=(const DSMCMonitor & that) {
        this->max_bin_v_inverse = that.max_bin_v_inverse;
        this->nbins = that.nbins;
        v_bin = new int[nbins];
        memcpy(v_bin, that.v_bin, sizeof(int)*nbins);
        return *this;
    }

    inline ~DSMCMonitor() {
        if (nbins && v_bin) delete[] v_bin;
        nbins = 0;
        v_bin = NULL;
    }

    DSMCData<_Particle> d;

  private:
    unsigned int nbins;
    double max_bin_v_inverse;
  public:
    inline const unsigned int & nBins() const { return  nbins; }
    int * v_bin;

    inline void binv(const double & v) {
        register unsigned int bin = (unsigned int)(v * max_bin_v_inverse * nbins);
        v_bin[bin < nbins ? bin : (nbins-1)]++;
    }

    /** startMeasure is prescribed for use in OctreeMonitor. */
    template <class T>
    inline void startMeasure(const T & o, const int & ptype) {
        d.zeroFields();

        /* set the position of this data point to the center of mass of the
         * particles in the octree node.  Not that it really matters.  This is
         * really only for debugging purposes.
         */
        o.copyRcm(d.pos);

        /* FIXME:  If ever multi-species runs are to be done, then we should
         * figure out how to handle this. */
        d.particle_type = ptype;
    }

    /** addMeasure is prescribed for use in OctreeMonitor. */
    inline void addMeasure(const _Particle & p) {
        d.data.name.N++;
        d.data.name.avgv[X] += p.vel[X];
        d.data.name.avgv[Y] += p.vel[Y];
        d.data.name.avgv[Z] += p.vel[Z];
        /* we'll use temperature[X:Z] as our temporary storage of v2 */
        register double vx2 = SQR(p.vel[X]),
                        vy2 = SQR(p.vel[Y]),
                        vz2 = SQR(p.vel[Z]);
        d.data.name.temperature[X]  += vx2;
        d.data.name.temperature[Y]  += vy2;
        d.data.name.temperature[Z]  += vz2;
        binv( sqrt( vx2 + vy2 + vz2 ) );
    }

    /** finishMeasure is prescribed for use in OctreeMonitor. */
    template <class T>
    inline void finishMeasure(const T & o) {
        /* calculate temperature and finish avgv calc. */
        if (!d.data.name.N) return;

        double OneOverN = 1.0 / d.data.name.N;

        for (int i = X; i <=Z; i++) {
            d.data.name.avgv[i] *= OneOverN;
            d.data.name.temperature[i] *= OneOverN;
            d.data.name.temperature[i] = particle_info[d.particle_type].mass
                                     * (d.data.name.temperature[i] - SQR(d.data.name.avgv[i]))
                                     / physical::constant::K_B;
        }

        /* compute number density. */
        d.data.name.numberDensity = d.data.name.N / o.getVolume();

        /* come up with a maximum relative velocity for the Particle/s contained
         * in this node.
         * */
        d.data.name.max_sigma_relspeed =
              particle_info[d.particle_type].cross_section
            * sqrt( /* 3* */physical::constant::K_B
                    * (
                         d.data.name.temperature[X]
                       + d.data.name.temperature[Y]
                       + d.data.name.temperature[Z]
                      )/* /3 */
                    / particle_info[d.particle_type].mass
             );
    }/* finishMeasure */

};

#endif // DSMCMONITOR_H
