// -*- c++ -*-
// $Id$
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log$
 *
 */

#ifndef FIELDOCTREECALC_H
#define FIELDOCTREECALC_H


#include <iostream>
#include <stdlib.h>
#include <string>
#include "common.h"
#include <olson-tools/Vector.h>

namespace FieldCalc {

    typedef struct {
        Vector<double,3> center;
        Vector<double,3> dx;
    } node_volume ;

    /** Samples a field according to it second derivate.
     * In other words, the density of sample points in a given region is
     * proportionally related to the second derivates.
     */
    template <class _FieldData, class Field>
    _FieldData * calculateField(const Field & field,
                                const node_volume & vol,
                                const double & min_absolute_error,
                                const int & nsamples,
                                int & nodes_remaining) {
        /* we first need to check whether this node is small enough for the
         * local field curvature. */

        /* start by finding the gradient in the center. */
        Vector<double,3> dF;
        field.grad(vol.center, dF);

        double err = 0;
        /* sample a view  gradients in the region. */
        for (int i = 0; i < nsamples; i++) {
            Vector<double,3> x_h;
            Vector<double,3> dFtest;
            for (int i = X; i <= Z; i++) {
                x_h[i] = vol.center[i] + vol.dx[i] * (2.0*MTRNGrand() - 1.0);
            }/* for */

            field.grad(x_h, dFtest);

            err += (dF - dFtest).abs() /* (dF.abs() + 1e-100) */;
        }/* for */

        if ((err/double(nsamples)) <= min_absolute_error) {
            /* this is a leaf node, so calculate the field for the center
             * point and return it. */
             Vector<double,3> F((const double[3]){0,0,0});
             field.field(vol.center, F);

            _FieldData * retval = new _FieldData(vol.center, F);
            nodes_remaining--;

            return retval;
        } else {
            /* this is not a leaf node, so we recurse down the eight children
             * of this node. */
            _FieldData * retval = NULL, ** retvalAddr = &retval;

            node_volume cvol;
            cvol.dx = vol.dx;
            cvol.dx *= 0.5;
            cvol.center[X] = vol.center[X] - cvol.dx[X];
            for ( int i = 0; i < 2;        i++, cvol.center[X] += vol.dx[X] ) {
                cvol.center[Y] = vol.center[Y] - cvol.dx[Y];
                for (int j = 0; j < 2;     j++, cvol.center[Y] += vol.dx[Y] ) {
                    cvol.center[Z] = vol.center[Z] - cvol.dx[Z];
                    for (int k = 0; k < 2; k++, cvol.center[Z] += vol.dx[Z] ) {
                        /* and now recurse with the corect volume. */
                        (*retvalAddr) =
                            calculateField<_FieldData>(field,
                                                       cvol,
                                                       min_absolute_error,
                                                       nsamples,
                                                       nodes_remaining);
                        if (nodes_remaining <= 0) return retval;

                        /* skip passed any linked lists returned. */
                        while ((*retvalAddr) != NULL)
                            retvalAddr = &((*retvalAddr)->next);

                        if (vol.dx[Z] == 0.0) break;
                    }/* for */
                    if (vol.dx[Y] == 0.0) break;
                }/* for */
                if (vol.dx[X] == 0.0) break;
            }/* for */

            /* and now return the complete linked list. */
            return retval;
        }/* else */
    }
}/* namespace */

#endif // FIELDOCTREECALC_H
