// -*- c++ -*-
// $Id: Octree.C,v 1.4 2005/06/07 10:00:08 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: Octree.C,v $
 * Revision 1.4  2005/06/07 10:00:08  olsonse
 * Many changes over the course of time
 *
 * Revision 1.3  2005/04/27 18:32:42  olsonse
 * Added some compile ifdefs for SunOS as well as PowerPC (OS X).
 *
 * Revision 1.2  2005/01/19 11:37:30  olsonse
 * Worked on automatic dimension testing again.  Seems to work for 1d now.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

/* Defines the octree node
 * See the header file (octree.h) for how space is subdivided to create an
 * octree.
 * */

#include "Octree.h"
#include <sstream>
#include <fstream>
#include <stdexcept>
#include "common.h"
#include "dsmc-config.h"
#include <olson-tools/power.h>
#include <olson-tools/logger.h>
#include <olson-tools/strutil.h>

#  if defined(DSMC_USE_PTHREAD)
#    include "PThreadBuildOctree.h"
#  endif

namespace octree {

/* TODO:  Pick some good value for the mem-limit */
REDEF_NEW_DEL_C_TEMPLATE(OctreeTemplate, OctreeClass,/* freed mem limit (instances) = */ 100000);

OctreeTemplate
int OctreeClass::maximum_depth = 10000;

OctreeTemplate
double OctreeClass::balance_criterion = 1.5;

OctreeTemplate
int OctreeClass::default_dimensions = 3;

template <class _Particle>
inline int partition( _Particle *a[],
                      const int & low,
                      const int & high,
                      const Vector<double,3> & p,
                      const int & val_indx );

template <class _Particle>
inline void SWAP(_Particle *a[], const int & left, const int & right);


/** A simple function to help debugging null pointers in the particle
 * list.
 * This function is a nice utility for using inside the debugger.
 *
 * @return The number of problems (unique to a particle or particle pointer).
 */
template <class _Particle>
int check_particle_array(const _Particle *particles[],
                          const int & low,
                          const int & high,
                          const LogLevel_t loglevel = LLOG_SEVERE) {
    int nulls = 0,
        nans  = 0,
        infs  = 0,
        snorms= 0,
        unique_problems = 0;

    for (int i = low; i<=high; i++) { 
        if (particles[i] == NULL) {
            unique_problems++;
            nulls++;
        } else {
            bool problem = false;
            for (int j = 0; j < 6; j++) {
                switch(fpclassify(particles[i]->pos[j])) {
                    case FP_NAN:
                        problem = true;
                        nans++;
                        break;
                    case FP_INFINITE:
                        problem = true;
                        infs++;
                        break;
                    case FP_SUBNORMAL:
                        problem = true;
                        snorms++;
                        break;
                    default:
                        break;
                }/* switch */
            }/* for */
            unique_problems += (int)problem;
        }/* else */
    }/* for */


    std::ostringstream err;
    if (nulls || nans || infs || snorms)
                    err << "Of " << (high-low+1) << " particles,  "
                                 " unique probs: " << unique_problems;
    if (nulls)      err << ",  nulls = " << nulls;
    if (nans)       err << ",  nans = " << nans;
    if (infs)       err << ",  infs = " << infs;
    if (snorms)     err << ",  snorms = " << snorms;
    if (nulls || nans || infs || snorms) {
        log_error(loglevel, 0, "%s\n", err.str().c_str());
        fflush(stderr);
    }

    return unique_problems;
}


OctreeTemplate
OctreeClass::Octree(int l, int h, int d): low(l), high(h),
                                     depth(d), leaves(0),
				     parent(NULL),
				     thisHasChildren(false),
				     findRcmFinished(false),
                                     vol_radius(0),
                                     dimensions(0),
				     marked(false),
                                     sibling_index(-1),
                                     extraFields() {
    memset(child, 0, 8*sizeof (void*));
    memset(&volb, 0, sizeof (vol_bounds));
    Rcm = 0.0;
}

OctreeTemplate
OctreeClass::~Octree() {
    for (int i = 0; i < 8; i++) {
	if (child[i]) {
            delete child[i];
            child[i] = NULL;
        }
    }
}

/**
 * findRcmAndVol
 * measure the center of mass and discover the minimum bounding box for the
 * particles held by this Octree.
 */
OctreeTemplate
void OctreeClass::findRcmAndVol( _Particle *a[], bool recursive /* = false */) {
    if (recursive && this->thisHasChildren) {
	for (int i = 0; i < 8; i++) {
	    if (child[i]) {
		child[i]->findRcmAndVol(a, true);
	    }//if
	}//for
    }

    /* call the functor that measures Rcm and bounding volume. */
    CMAndVolMeasurement::measure(*this, a);

    /* detect dimensionality. */
    double xlen = (volb.max[X] - volb.min[X]),
           ylen = (volb.max[Y] - volb.min[Y]),
           zlen = (volb.max[Z] - volb.min[Z]),
           vol = 1;
    dimensions = 0;

    if (xlen) { vol*=xlen; dimensions++; }
    if (ylen) { vol*=ylen; dimensions++; }
    if (zlen) { vol*=zlen; dimensions++; }

    switch (dimensions) {
        case 3: /* 3d */
            /* (3/(4 pi))^(1/3) == 0.620350490899400 */
            vol_radius = 0.620350490899400 * fast_pow(vol, 0.333333333333333);
            break;
        case 2: /* 2d */
            /* sqrt(1/pi) == 0.564189583547756 */
            vol_radius =  0.564189583547756 * sqrt(vol);
            break;
        case 1: /* 1d */
            vol_radius = vol;
            break;
        default: /* 0 d??? */ {
            /* In this case, we'll attempt to inherit dimensionality from the
             * parent node.  Otherwise, we fail?  Also, we attempt to set the
             * volume by calling the particles "effective_radius" function.
             *
             * FIXME:  It would probably be better to first try and do an
             * average of the sibling nodes' volumes.
             */
            vol_radius = a[low]->effective_radius();

            if (parent) dimensions = parent->dimensions;
            else {
                dimensions = default_dimensions;
                if (numberParticles() > 1) {
                    log_warning(
                            FILE_LINE "Octree node dimensions error.  "
                                      "contains particles in zero-D root node:  "
                                      "[%d - %d]\n", low, high
                               );
                    log_warning("default dimension (%d) used\n", default_dimensions);
                }
            }
        }
    }/* switch */

    /* give a warning if the dimensionality of the problem has changed. */
    if (parent && parent->dimensions != dimensions) {
        log_warning("Octree dimensions changed [%d - %d]:  parent(%d) -> this(%d) \n",
                    low, high, parent->dimensions, dimensions);
    }

    findRcmFinished = true;
}

#define OMP_SHARED shared(minp,minp2X,dx,splitY,a,dxbc,children)
#define OMP_PRIVATE private(pivot)
#define OMP_COMMON default(none) OMP_PRIVATE OMP_SHARED
#define OMP_NESTED shared(splitZ)

#define OMP_LOOP
//#define OMP_SECTIONS
#ifndef OMP_SECTIONS
#  undef DSMC_NESTED_OMP_SECTION
#endif

/** simple test function for buildOctree.
 * I'll use an inline function here so that I don't have to evaluate things
 * like child[0].low a lot.  Very small speed up, but also easier to code and
 * read.
 */
inline bool enoughRemainingParticles(const double & minp, const int & pivot, const int & low, const int & high ) {
    return ( (pivot - low) >= minp && (high + 1 - pivot) >= minp );
}

//#define DEBUGBUILD
/* build Octree using a quick sort algorithm
 * returns the number of leaf nodes in the octree just created.
 * NOTE:  by convention partition returns an item to the "right" of the pivot
 * point.
 */
OctreeTemplate
int OctreeClass::buildOctree( _Particle *a[],
                              const LeafModel & leafModel ) {
    int pivot;
    int children = 0; /* this is just for zero-d error checking. */

    /* first start by measuring the location of the center of mass and minimum
     * bounding volume for this collection of particles
     */

    findRcmAndVol(a);

    double minp = leafModel.maxParticlesPerLeaf(*this);
    double minp2X = 2*minp;

    if ( numberParticles() >= minp2X ) {
	if ( depth > Octree::maximum_depth ) {
	    /* our maximum depth is set quite high, so this is a real
	     * indicator that something is wrong
             * NOTE:  with the new subdivide recovery code below, this really
             * should never be hit again.
             */
	    log_severe(FILE_LINE "octree depth == %d,  runaway condition\n", depth);
            log_severe("%i-D, lo: %i, hi: %i, radius: %lg, Rcm=(%lg,%lg,%lg), "
                       "volb=(%lg->%lg, %lg->%lg, %lg->%lg)\n", 
                       dimensions, low, high,
                       vol_radius, Rcm[0], Rcm[1], Rcm[2],
                       volb.min[0], volb.max[0],
                       volb.min[1], volb.max[1],
                       volb.min[2], volb.max[2]
                      );
            fflush(stderr);

            /* let's do a check on all of the particles, just to see if they
             * have any problems.  At least we should be able to report
             * problems. */
            check_particle_array((const _Particle**)a, low, high);

            /* we will dump the problem particles and then throw an error.
             * But, only dump the particles if the user allows it [Default no].
             * We'll also only throw the error if allowed.  That way a
             * debug session can keep going; a normal program will die soon of
             * stack fault anyway.
             */
            const char * DUMP = getenv("OCTREE_DUMP_RUNAWAY");

            if (DUMP) {
                std::ofstream fout(DUMP);

                for (int i = low; i <= high; i++) {
                    fout << (*a[i]) << '\n';
                }
                fout.close();

                /* we'll sleep for a tiny bit to allow for buffers
                 * to flush (needed for MPI) */
                sleep(5); 
                THROW(std::runtime_error,FILE_LINE "no use continuing on");
            }
	}

	thisHasChildren = true;
        double dx[3] = {volb.dx(X), volb.dx(Y), volb.dx(Z)};
        double dxbc[3] = {balance_criterion*dx[X],
                          balance_criterion*dx[Y],
                          balance_criterion*dx[Z]};

	/* start with partitioning along x-direction */
        if (dxbc[X] >= dx[Y] && dxbc[X] >= dx[Z]) {
            pivot = partition( a, low, high, Rcm, X );

            if (!enoughRemainingParticles(minp,pivot,low,high)) {
                /* all of the particles belong in one of these nodes */
                child[0] = new Octree(low, high, depth+1); children++;
                child[4] = NULL;
            } else {
                /* particles found in both x-partitions */
                child[0] = new Octree(low, pivot-1, depth+1); children++;
                child[4] = new Octree(pivot, high, depth+1); children++;

                /* rescale dx[X] */
                dx[X] *= 0.5;
            }
        } else {
            /* no split in X direction */
            child[0] = new Octree(low, high, depth+1); children++;
            child[4] = NULL;
        }

        bool splitY = dxbc[Y] >= dx[X] && dxbc[Y] >= dx[Z];

#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp parallel sections OMP_COMMON
{ /* parallel sections */
#  pragma omp section
    { /* y-direction (1) section */
#endif
	/* continue now with partitioning along y-direction if needed */
	if ( child[0]->numberParticles() >= minp2X ) {
            double dy = dx[Y];
            if (splitY) {
                pivot = partition( a, child[0]->low, child[0]->high, Rcm, Y );

                if (enoughRemainingParticles(minp,pivot,child[0]->low,child[0]->high)) {
                    /* both nodes are needed, so divide the particles */
                    child[3] = new Octree(pivot, child[0]->high, depth+1);
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                    children++;
                    child[0]->high = pivot-1;

                    /* rescale dx[Y] */
                    dy *= 0.5;
                } else {
                    child[3] = NULL;
                }
            } else {
                /* no split in Y direction */
                child[3] = NULL;
            }

            bool splitZ = dxbc[Z] >= dx[X] && dxbc[Z] >= dy;

#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
#  pragma omp parallel sections OMP_COMMON OMP_NESTED
        { /* nested parallel sections */
#  pragma omp section
            { /* z-direction (1:1) section */
#endif
	    /* and finally doing the z-direction */
	    if ( child[0]->numberParticles() >= minp2X ) {
                if (splitZ) {
                    pivot = partition( a, child[0]->low, child[0]->high, Rcm, Z );

                    if (enoughRemainingParticles(minp,pivot,child[0]->low,child[0]->high)) {
                        /* both nodes are needed, so divide the particles */
                        child[1] = new Octree(pivot, child[0]->high, depth+1);
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                        children++;
                        child[0]->high = pivot-1;
                    } else {
                        child[1] = NULL;
                    }
                } else {
                    /* no split in Z direction */
                    child[1] = NULL;
                }
	    }//if


#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
            } /* z-direction (1:1) section */
#  pragma omp section
            { /* z-direction (1:2) section */
#endif
	    /* and finally doing the z-direction */
	    if ( child[3] && child[3]->numberParticles() >= minp2X ) {
                if (splitZ) {
                    pivot = partition( a, child[3]->low, child[3]->high, Rcm, Z );

                    if (enoughRemainingParticles(minp,pivot,child[3]->low,child[3]->high)) {
                        /* both nodes are needed, so divide the particles */
                        child[2] = new Octree(pivot, child[3]->high, depth+1);
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                        children++;
                        child[3]->high = pivot-1;
                    } else {
                        child[2] = NULL;
                    }
                } else {
                    /* no split in Z direction */
                    child[2] = NULL;
                }
	    }//if

#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
            } /* z-direction (1:2) section */
        } /* nested parallel sections */
#endif
	}//if


#if defined(_OPENMP) && defined(OMP_SECTIONS)
    } /* y-direction (1) section */
#  pragma omp section
    { /* y-direction (2) section */
#endif
	/* continue now with partitioning along y-direction if needed */
	if ( child[4] && child[4]->numberParticles() >= minp2X ) {
            double dy = dx[Y];
            if (splitY) {
                pivot = partition( a, child[4]->low, child[4]->high, Rcm, Y );

                if (enoughRemainingParticles(minp,pivot,child[4]->low,child[4]->high)) {
                    /* both nodes are needed, so divide the particles */
                    child[7] = new Octree(pivot, child[4]->high, depth+1); 
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                    children++;
                    child[4]->high = pivot-1;

                    /* rescale dx[Y] */
                    dy *= 0.5;
                } else {
                    child[7] = NULL;
                }
            } else {
                /* no split in Y direction */
                child[7] = NULL;
            }

            bool splitZ = dxbc[Z] >= dx[X] && dxbc[Z] >= dy;


#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
#  pragma omp parallel sections OMP_COMMON OMP_NESTED
        { /* nested parallel sections */
#  pragma omp section
            { /* z-direction (2:1) section */
#endif
	    /* and finally doing the z-direction */
	    if ( child[4]->numberParticles() >= minp2X ) {
                if (splitZ) {
                    pivot = partition( a, child[4]->low, child[4]->high, Rcm, Z );

                    if (enoughRemainingParticles(minp,pivot,child[4]->low,child[4]->high)) {
                        /* both nodes are needed, so divide the particles */
                        child[5] = new Octree(pivot, child[4]->high, depth+1); 
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                        children++;
                        child[4]->high = pivot-1;
                    } else {
                        child[5] = NULL;
                    }
                } else {
                    /* no split in Z direction */
                    child[5] = NULL;
                }
	    }//if

#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
            } /* z-direction (2:1) section */
#  pragma omp section
            { /* z-direction (2:2) section */
#endif
	    /* and finally doing the z-direction */
	    if ( child[7] && child[7]->numberParticles() >= minp2X ) {
                if (splitZ) {
                    pivot = partition( a, child[7]->low, child[7]->high, Rcm, Z );

                    if (enoughRemainingParticles(minp,pivot,child[7]->low,child[7]->high)) {
                        /* both nodes are needed, so divide the particles */
                        child[6] = new Octree(pivot, child[7]->high, depth+1);
#if defined(_OPENMP) && defined(OMP_SECTIONS)
#  pragma omp atomic
#endif
                        children++;
                        child[7]->high = pivot-1;
                    } else {
                        child[6] = NULL;
                    }
                } else {
                    /* no split in Z direction */
                    child[6] = NULL;
                }
	    }//if
#if defined(_OPENMP) && defined(DSMC_NESTED_OMP_SECTION)
            } /* z-direction (2:2) section */
        } /* nested parallel sections */
#endif
	}//if
#if defined(_OPENMP) && defined(OMP_SECTIONS)
    } /* y-direction (2) section */
} /* parallel sections */
#endif


        if (children > 1) {
            /* now tell the children to clean up their own rooms.
             * this will work when we have each node detect their own center of
             * mass and their own volume extrema.
             */
            leaves = 0;

#if (defined(_OPENMP) || defined(DSMC_USE_PTHREAD)) && defined(OMP_LOOP)

#  if !defined(DSMC_USE_PTHREAD)
#    define get_max_threads       omp_get_max_threads
#  endif

            if ( get_max_threads() <= 1 ||
                 depth > ((get_max_threads()-1)/(1<<dimensions))) {
                /* we aren't going to split off any more */
#endif

                for (int i = 0; i < 8; i++) {
                    if (child[i]) {
                        child[i]->parent = this;
                        child[i]->sibling_index = i;
                        leaves += child[i]->buildOctree(a, leafModel);
                    }//if
                }//for

#if (defined(_OPENMP) || defined(DSMC_USE_PTHREAD)) && defined(OMP_LOOP)
            } else {

#  if !defined(DSMC_USE_PTHREAD) /* prefer pthreads for divide and conquer */
    /* NOTE : There must be something fundamentally wrong with this OMP code.
     * Niobe really doesn't like when I try to create a whole bunch of OMP
     * tasks recursively.  Maybe we'll just have to remove this code.  */ 
#    pragma omp parallel for default(none) shared(a,min_part,leaves)
                for (int i = 0; i < 8; i++) {
                    if (child[i]) {
                        child[i]->parent = this;
                        child[i]->sibling_index = i;
#    pragma omp atomic
                        leaves += child[i]->buildOctree(a, leafModel);
                    }//if
                }//for

#  else /* DSMC_USE_PTHREAD is defined! */

                PThreadBuildOctree<Octree> builder;

                for (int i = 0; i < 8; i++) {
                    if (child[i]) {
                        child[i]->parent = this;
                        child[i]->sibling_index = i;
                        builder.buildOctree(i, child[i], a, leafModel);
                    }//if
                }//for
                leaves = builder.gatherLeaves();

#  endif /* DSMC_USE_PTHREAD */

            }
#endif // _OPENMP or DSMC_USE_PTHREAD and OMP_LOOP

        } else {
            /* we were supposed to divide the Octree node, but somehow can't.  
             * We will stop recursion here, delete the singly created child,
             * and declare ourselves a leafnode. 
             * If this is not just a case where
             * this.numberParticles() ~= minp2X (such that statistical
             * variations in the density distribution allow for this
             * unsophisticated attempt to split the node to fail) then this is
             * an error.  Perhaps we ought to test for such an error case and
             * emit a warning. 
             */

            for (int i = 0; i < 8; i++) {
                if (child[i]) {
                    delete child[i];
                    child[i] = NULL;
                    children--;
                }//if
            }//for

            if (children < 0) {
                /* this is yet another error condition! */
                log_severe("Octree::buildOctree(subdivide recovery):  "
                    "deleted more than one (%d) child!\n", (1-children)
                );
            }

            thisHasChildren = false;
            leaves = 1;
        }

        return leaves;
    } else {
        /* this only has itsself as a leafnode, so return 1 */
        return leaves=1;
    }
}

template <class _Particle>
inline int partition( _Particle *a[], const int & low, const int & high, const Vector<double,3> & p, const int & val_indx ) {
    int swaps = 0, left = low, right = high;
    while ( left < right ) {
	/* Move left while item < pivot */
	while( left < right && a[left]->pos[val_indx] <= p[val_indx] ) left++;

	/* Move right while item > pivot */
	while( left < right && a[right]->pos[val_indx] > p[val_indx] ) right--;

	if ( left < right ) {
            SWAP(a,left,right);
            swaps++;
        }
    }

    /* if there have been no swaps, then we need to determine whether it was
     * already sorted. */
    if (swaps ||
        (a[low]->pos[val_indx] <= p[val_indx] &&
         a[high]->pos[val_indx] > p[val_indx]
        )
       ) return right;
    else {
        /* didn't have to sort because everything was on one side of the
         * partition. */
        return low;
    }
}

template <class _Particle>
inline void SWAP(_Particle *a[], const int & left, const int & right) {
    _Particle * tmp = a[left];
    a[left] = a[right];
    a[right] = tmp;
}

OctreeTemplate
std::ostream & operator<<(std::ostream & output,const OctreeClass & o){
    char tabs[5];
    int len = o.depth<5 ? o.depth : 5;
    if (len) memset(tabs, '\t', len);
    tabs[len] = '\0';
    output
	<< tabs << "{\n"
	<< tabs << "\tdepth = " << o.depth << '\n'
	<< tabs << "\tnumber leaves = " << o.leaves << '\n'
	<< tabs << "\tnumber particles = " << o.numberParticles() << '\n'
	<< tabs << "\tlow = " << o.low << '\n'
	<< tabs << "\thigh = " << o.high << '\n'
	<< tabs << "\tRcm = ("<<o.Rcm[X]<<','<<o.Rcm[Y]<<','<<o.Rcm[Z]<<")\n"
	<< tabs << "\tvolume bounds = ("<<o.volb.min[X]<<','<<o.volb.min[Y]<<','<<o.volb.min[Z]<<") --> "
					<<o.volb.max[X]<<','<<o.volb.max[Y]<<','<<o.volb.max[Z]<<")\n"
	<< tabs << "\tvolume = " << o.getVolume() << '\n'
	<< tabs << "\tmarked = " << (o.marked? "true":"false") << '\n';
    for (int i = 0; i < 8; i++) {
	if (o.child[i]) output << tabs << "\tchild[" << i << "] = \n" << *o.child[i];
    }
    output
	<< tabs << "}\n";
    return output;
} // operator<< Octree

OctreeTemplate
bool OctreeClass::gnuplotLeavesVol(std::ostream & output, const bool & std_in, const int & max_depth) {
    if (!thisHasChildren || (max_depth >= 0 && depth >= max_depth)) {
	gnuplotVol(output, std_in);
        return thisHasChildren;
    } else {
        bool retval = false;
	for (int i = 0; i < 8; i++) {
	    if(child[i]) {
                retval |= child[i]->gnuplotLeavesVol(output, std_in, max_depth);
            }
	}
        return retval;
    }
}

OctreeTemplate
void OctreeClass::gnuplotVol(std::ostream & output, const bool & std_in) {
    if (marked) {
	output
	<< volb.min[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.max[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.min[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.max[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n';
    }
    output
	<< volb.min[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.max[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.max[X] <<' '<< volb.max[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.min[X] <<' '<< volb.max[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.min[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n'

	<< volb.min[X] <<' '<< volb.min[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.min[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.min[X] <<' '<< volb.max[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.min[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n'

	<< volb.max[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.max[X] <<' '<< volb.min[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.min[X] <<' '<< volb.min[Y] <<' '<< volb.max[Z] << '\n'
	<< volb.max[X] <<' '<< volb.min[Y] <<' '<< volb.max[Z] << '\n'

	<< volb.max[X] <<' '<< volb.min[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.max[X] <<' '<< volb.max[Y] <<' '<< volb.min[Z] << '\n'
	<< volb.max[X] <<' '<< volb.max[Y] <<' '<< volb.max[Z] << '\n';

    if (std_in) {
        output << "e\n";
    } else {
        output << "\n";
    }
}

OctreeTemplate
std::string & OctreeClass::createGnuplotLeaveSPLOT(
        std::string & str,
        const std::string & plotopt,
        const int leaf_count,
        const std::string & tree_dat) {
    std::ostringstream ostr;

    if (tree_dat.length() > 0) {
        for( int i = 0; i < leaf_count; i++ ) {
            ostr << " '"<< tree_dat << "' i " << i << ' ' << plotopt << ' ';
            if (i < leaf_count-1) ostr << ',';
        }
    } else {
        for( int i = 0; i < leaf_count; i++ ) {
            ostr << " '-' " << plotopt << ' ';
            if (i < leaf_count-1) ostr << ',';
        }
    }

    str = ostr.str();
    return str;
}

OctreeTemplate
OctreeClass & OctreeClass::findNearestLeaf(double x, double y, double z) {
    if (!thisHasChildren) return  *this;
    int child_index = 0;
    int sign = 1;

    /*else*/
    /* note from buildOctree that:
     * node  0 must exist for any other to exist
     * node  4 must exist if there was an x partition
     * node  3 must exist if 0 was y-partitioned
     * node  7 must exist if 4 was y-partitioned
     * and so on
                   0                    [initial distribution]
             0           4              [x partition]
          0     3     4     7           [y partition]
         0 1   2 3   4 5   6 7          [z partition]
     */
    if (child[4] && x > Rcm[X]) child_index+=4;

    /* test to see if node 3|7 exist */
    if (child[child_index+3]) {
	if (y > Rcm[Y]) child_index+=3, sign =-1;
    }

    /* test to see if 1|2|5|6 exist */
    if (child[child_index+sign] && z > Rcm[Z]) child_index+=sign;

    return child[child_index]->findNearestLeaf(x,y,z);
}



template <class OctreeType>
inline void measureCMAndVolShrink(OctreeType & o,
                               Vector<double,3> & Rcm,
                               vol_bounds & volb,
                               typename OctreeType::OParticle *a[]) {
    Rcm = 0.0;
    memcpy(volb.min, a[o.low]->pos, sizeof(volb.min));
    memcpy(volb.max, volb.min, sizeof(volb.min));

    /* loop through and sum up positions and auto-expand the bounding volume */
    for (int i = o.low; i <= o.high; i++) {
        Rcm += V3C(a[i]->pos);

        if (a[i]->pos[X] > volb.max[X]) volb.max[X] = a[i]->pos[X];
        else if (a[i]->pos[X] < volb.min[X]) volb.min[X] = a[i]->pos[X];

        if (a[i]->pos[Y] > volb.max[Y]) volb.max[Y] = a[i]->pos[Y];
        else if (a[i]->pos[Y] < volb.min[Y]) volb.min[Y] = a[i]->pos[Y];

        if (a[i]->pos[Z] > volb.max[Z]) volb.max[Z] = a[i]->pos[Z];
        else if (a[i]->pos[Z] < volb.min[Z]) volb.min[Z] = a[i]->pos[Z];
    }

    Rcm /= o.numberParticles();
}


template <class OctreeType>
inline void CMAndVolShrink::measure(OctreeType & o, typename OctreeType::OParticle *a[]) {
    measureCMAndVolShrink(o, o.Rcm, o.volb, a);
}

template <class OctreeType>
inline void CMAndVolNotShrink::measure(OctreeType & o, typename OctreeType::OParticle *a[]) {
    if (o.parent == NULL) {
        /* we allow the parent node to shrink. */
        measureCMAndVolShrink(o, o.Rcm, o.volb, a);
        return;
    }

    /* now determined the center of mass and the shrunken volume. */
    vol_bounds volb;
    measureCMAndVolShrink(o, o.Rcm, volb, a);


    /* calculate the bounding volume. */
    o.volb = o.parent->volb;

    /* @see Octree.h for a description of the filling order. */
    switch(o.sibling_index) {
        case 0:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            if (o.parent->child[3]) o.volb.max[Y] = o.parent->Rcm[Y];
            if (o.parent->child[1]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 1:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            if (o.parent->child[3]) o.volb.max[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 2:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 3:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            if (o.parent->child[2]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 4:
            o.volb.min[X] = o.parent->Rcm[X];
            if (o.parent->child[7]) o.volb.max[Y] = o.parent->Rcm[Y];
            if (o.parent->child[5]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 5:
            o.volb.min[X] = o.parent->Rcm[X];
            if (o.parent->child[7]) o.volb.max[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 6:
            o.volb.min[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 7:
            o.volb.min[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            if (o.parent->child[6])  o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case -1:
        default:
            THROW(std::runtime_error,FILE_LINE "invalid sibling index for non-root node!");
    }/* switch */

    /* now perform some sanity checks on the volume sizes to see if we really
     * should shrink one or more of the dimensions. */
    const double FComp = 1.0;

    if ( (volb.min[X] - o.volb.min[X]) > (volb.dx(X) * FComp) ) {
        o.volb.min[X] = volb.min[X];
    }

    if ( (o.volb.max[X] - volb.max[X]) > (volb.dx(X) * FComp) ) {
        o.volb.max[X] = volb.max[X];
    }

    if ( (volb.min[Y] - o.volb.min[Y]) > (volb.dx(Y) * FComp) ) {
        o.volb.min[Y] = volb.min[Y];
    }

    if ( (o.volb.max[Y] - volb.max[Y]) > (volb.dx(Y) * FComp) ) {
        o.volb.max[Y] = volb.max[Y];
    }

    if ( (volb.min[Z] - o.volb.min[Z]) > (volb.dx(Z) * FComp) ) {
        o.volb.min[Z] = volb.min[Z];
    }

    if ( (o.volb.max[Z] - volb.max[Z]) > (volb.dx(Z) * FComp) ) {
        o.volb.max[Z] = volb.max[Z];
    }

}

template <class OctreeType>
inline void GeometricDivision::measure(OctreeType & o, typename OctreeType::OParticle *a[]) {
    if (o.parent == NULL) {
        /* we allow the parent node to shrink. */
        measureCMAndVolShrink(o, o.Rcm, o.volb, a);

        /* now we need to set Rcm to the geometric center of the bounded
         * volume. */
        o.Rcm[X] = 0.5 * (o.volb.max[X] + o.volb.min[X]);
        o.Rcm[Y] = 0.5 * (o.volb.max[Y] + o.volb.min[Y]);
        o.Rcm[Z] = 0.5 * (o.volb.max[Z] + o.volb.min[Z]);
        return;
    }

    /* calculate the bounding volume. */
    o.volb = o.parent->volb;

    /* @see Octree.h for a description of the filling order. */
    switch(o.sibling_index) {
        case 0:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            if (o.parent->child[3]) o.volb.max[Y] = o.parent->Rcm[Y];
            if (o.parent->child[1]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 1:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            if (o.parent->child[3]) o.volb.max[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 2:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 3:
            if (o.parent->child[4]) o.volb.max[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            if (o.parent->child[2]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 4:
            o.volb.min[X] = o.parent->Rcm[X];
            if (o.parent->child[7]) o.volb.max[Y] = o.parent->Rcm[Y];
            if (o.parent->child[5]) o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case 5:
            o.volb.min[X] = o.parent->Rcm[X];
            if (o.parent->child[7]) o.volb.max[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 6:
            o.volb.min[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            o.volb.min[Z] = o.parent->Rcm[Z];
            break;

        case 7:
            o.volb.min[X] = o.parent->Rcm[X];
            o.volb.min[Y] = o.parent->Rcm[Y];
            if (o.parent->child[6])  o.volb.max[Z] = o.parent->Rcm[Z];
            break;

        case -1:
        default:
            THROW(std::runtime_error,FILE_LINE "invalid sibling index for non-root node!");
            break;
    }/* switch */

    /* calculate the Rcm point. */
    o.Rcm[X] = o.volb.min[X] + 0.5*o.volb.dx(X);
    o.Rcm[Y] = o.volb.min[Y] + 0.5*o.volb.dx(Y);
    o.Rcm[Z] = o.volb.min[Z] + 0.5*o.volb.dx(Z);
}

std::ostream & operator<<(std::ostream & output, const UniformLeafModel & ulm) {
    output << ulm.max_particles_per_leaf;
    return output;
}

}

