// -*- c++ -*-
// $Id: Octree.h,v 1.3 2005/06/07 10:00:09 olsonse Exp $
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log: Octree.h,v $
 * Revision 1.3  2005/06/07 10:00:09  olsonse
 * Many changes over the course of time
 *
 * Revision 1.2  2005/01/19 11:37:30  olsonse
 * Worked on automatic dimension testing again.  Seems to work for 1d now.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#ifndef OCTREE_H
#define OCTREE_H


/** \file
 * The Octree class defines the octree node:
 * The picture below describes the method of subdividing one node into its
 * sub-octree consisting of 8 child-nodes.
 * The numbers indicate the indices in the child array.  Because we use
 * this set of indices, the x-plane test must be done first, followed by the
 * y-plane, and then finally the z-plane tests.
 \verbatim
     _________________
    /  2    /   6   / |
   /_______/_______/  |       Z     Y
  /       /       /|6 |       |    /
 /   1   /    5  / | /|       |   /
|_______________/  |/ |       |  /
|       |       |5 |  |       | /
|   1   |   5   | /|7 |       |__________X
|       |       |/ | /
|-------|-------|  |/
|       |       |4 |
|   0   |   4   | /
|_______|_______|/
\endverbatim

 * Children being used actually depends on how many particles there are in the
 * node as well as how these are organized.  We do not keep empty children, so
 * if there are too few particles, the node will have less than 8 children.
 *
 * The following describes the order of filling the chidren nodes.
 * - (0)  [Initial Distribution]
 *      - (0) splits into: (0) & (4)  [x-plane test]
 *              - (0) splits into: (0) & (3)  [y-plane test]
 *                      - (0) splits into: (0) & (1)  [z-plane test]
 *                      - (3) splits into: (3) & (2)  [z-plane test]
 *                      .
 *              - (4) splits into: (4) & (7)  [y-plane test]
 *                      - (4) splits into: (4) & (5)  [z-plane test]
 *                      - (7) splits into: (7) & (6)  [z-plane test]
 *                      .
 *              .
 *      .
 * .
 *
 * Or using a tree view of the same filling order:
 \verbatim
                   0                    [initial distribution]
             0           4              [x partition]
          0     3     4     7           [y partition]
         0 1   2 3   4 5   6 7          [z partition]
 \endverbatim
*/

/** \example testbasic.C
 * A very basic example of sorting particles into an octree.
 * Particles are first added to an OctreeContainer, after which the Octree is
 * built.  The execution time is saved to a file, after which the program
 * begins again with prompting the user to specify the number of particles to
 * include in the build. 
 *
 * At runtime, the user is asked to specify a number of particles to include
 * in the build.  The user is also asked to specify the rough number of
 * particles per Octree node.  This code currently only uses the
 * UniformLeafModel to determine the number of particles per each leaf node.  
 */

/** \example testtiming.C
 * Very similar to the testbasic.C example, except that the user is not
 * prompted for anything.  Rather, the code loops through increasing
 * numbers of total particles in the sort and also loops through increasing
 * numbers of particles in the leaf nodes.  
 */


#include <iostream>
#include <stdlib.h>
#include <string>
#include <math.h>
#include <iterator>
#include <stdexcept>
#include "common.h"

#include <olson-tools/memory.h>
#include <olson-tools/logger.h>
#include <olson-tools/Vector.h>
#include <olson-tools/power.h>
#include <olson-tools/ompexcept.h>

/** Collection of Octree utilities.
*/
namespace octree {

/** Simple class with extrema values of an octree node's bounding volume. 
 * By simple, I mean pretty simple.  This means, for example, that although
 * there is no error checking against specifying min[] values that are greater
 * than max[] values, doing so will have ill-defined effects on whatever other
 * code that depends on this.   
 */
class vol_bounds {
  public:
    double max[3]; /* Maximum extents of bounded volume. */
    double min[3]; /* Minimum extents of bounded volume. */

    /** Default vol_bounds constructor initializes the volume to zero. */
    inline vol_bounds() {
        memset(max,0,sizeof(max));
        memset(min,0,sizeof(min));
    }

    /** vol_bounds copy constructor. */
    inline vol_bounds(const vol_bounds & that) {
        *this = that;
    }

    /** Functional access to ith maximum extent of boundary--non-const.  This
     * was add primarily to satisfy the interface needed for ParticleNode. 
     */
    inline double & Max(const int & i) { return max[i]; }

    /** Functional access to ith maximum extent of boundary--const.  This
     * was add primarily to satisfy the interface needed for ParticleNode. 
     */
    inline const double & Max(const int & i) const { return max[i]; }

    /** Functional access to ith minimum extent of boundary--non-const.  This
     * was add primarily to satisfy the interface needed for ParticleNode. 
     */
    inline double & Min(const int & i) { return min[i]; }

    /** Functional access to ith minimum extent of boundary--const.  This
     * was add primarily to satisfy the interface needed for ParticleNode. 
     */
    inline const double & Min(const int & i) const { return min[i]; }

    /** Computes max[i] - min[i].
     * DOES NOT use fabs().
     */
    inline double dx(const int & i) { return max[i] - min[i]; }

    /** Determine whether the specified 3-dimensional location is bounded by
     * this volume.
     * @param x
     *     3-vector specifying location to test for.
     */
    inline bool contains(const double x[] /* 3-vector */) const {
        return (max[X] >= x[X] && x[X] >= min[X] &&
                max[Y] >= x[Y] && x[Y] >= min[Y] &&
                max[Z] >= x[Z] && x[Z] >= min[Z]);
    }

    /** Determine whether the specified 3-dimensional location is bounded by
     * this volume, within the specified tolerance.
     * @param x
     *     3-vector specifying location to test for.
     * @param tolerance
     *     Tolerance to use when testing for boundedness (better be a positive
     *     quantity).
     */
    inline bool contains(const double x[] /* 3-vector */, const double & tolerance) const {
        return ((max[X]+tolerance) >= x[X] && x[X] >= (min[X]-tolerance) &&
                (max[Y]+tolerance) >= x[Y] && x[Y] >= (min[Y]-tolerance) &&
                (max[Z]+tolerance) >= x[Z] && x[Z] >= (min[Z]-tolerance));
    }

    /**  Tests for overlap between two given bounded volumes. */
    inline bool overlaps(const vol_bounds & that) const {
        return
        (min[X] <= that.Max(X) && max[X] >= that.Min(X)) &&
        (min[Y] <= that.Max(Y) && max[Y] >= that.Min(Y)) &&
        (min[Z] <= that.Max(Z) && max[Z] >= that.Min(Z));
    }

    /** Assignment operator. */
    inline const vol_bounds & operator=(const vol_bounds & that) {
        memcpy(max, that.max, sizeof(max));
        memcpy(min, that.min, sizeof(min));
        return *this;
    }

    /** Output stream writer. */
    std::ostream & print(std::ostream & output) const {
        std::copy(max, min+3, std::ostream_iterator<double>(output, "\t"));
        return output;
    }
};

/** A leaf model that produces a (roughly) uniform particle distribution per
 * leaf. 
 */
class UniformLeafModel {
  public:
    /** Rough number of particles per leaf node.
     * Actually, on average, the number of particles per leaf node will be
     * (3/2) * max_particles_per_leaf.
     */
    double max_particles_per_leaf;

    /** Default constructor--defaults 60(\f$ \approx 90 \f$) particles per leaf. */
    UniformLeafModel(const double & n = 60.0) : max_particles_per_leaf(n) {}

    /** Functional access to max_particles_per_leaf--const.  This function
     * satisfies the interface defined by the Octree class. */
    template <class _Octree>
    const double & maxParticlesPerLeaf(const _Octree & o) const {
        return max_particles_per_leaf;
    }
};

std::ostream & operator<<(std::ostream & output,const UniformLeafModel & ulm);

/** Just a default particle class that allows easier use of static functions.
 * Real Octree use must of course be with a real particle definition. */
class NotAParticle {};

#define OctreeTemplate  template<class _Particle, class _extraFields, class CMAndVolMeasurement, class LeafModel>
#define DefaultOctreeTemplate template <class _Particle = NotAParticle, class _extraFields = nothing, class CMAndVolMeasurement = CMAndVolShrink, class LeafModel = UniformLeafModel>
#define OctreeTemplateArgs _Particle,_extraFields,CMAndVolMeasurement,LeafModel
#define OctreeClass     Octree<OctreeTemplateArgs>

OctreeTemplate class Octree;

OctreeTemplate
std::ostream & operator<<(std::ostream & output,const OctreeClass & o);

/** Just what it says:  nothing. */
typedef struct {} nothing;

/** Center of mass division model of Octree that shrinks nodes to their
 * minimum size.
 */
class CMAndVolShrink {
  public:
    template <class OctreeType>
    static inline void measure(OctreeType & o, typename OctreeType::OParticle *a[]);
};

/** Center of mass division model of Octree that only shrinks nodes when they
 * have significant empty space.
 */
class CMAndVolNotShrink {
  public:
    template <class OctreeType>
    static inline void measure(OctreeType & o, typename OctreeType::OParticle *a[]);
};

/** Division model of Octree that uses the geometric center of nodes as the
 * pivot point.
 */
class GeometricDivision {
  public:
    template <class OctreeType>
    static inline void measure(OctreeType & o, typename OctreeType::OParticle *a[]);
};
    
/** Implementation of an octree.
 * @param _Particle
 *     The particle class to sort.
 *
 * @param _extraFields
 *     Auxillary struct/class storage space to include in each node for
 *     specialization.
 *     [Default nothing]
 *
 * @param CMAndVolMeasurement
 *     The class that will provide the                                  <br>
 *  template <class OctreeType>                                         <br>
 *  static inline void measure(OctreeType & o, typename OctreeType::OParticle *a[]);         <br>
 *     function.  This function measures/calculates/sets the center of the
 *     node (Octree::Rcm) as well as the bounding volume (Octree::volb).
 *     The inline is of course optional, but recommended.
 *     This function will obtain friend privileges to the Octree node.  This
 *     function should use this friendship well and carefully.
 *     [Default  CMAndVolShrink]
 * @param LeafModel
 *     Sets the model for the leaf nodes of the tree.  The model determines
 *     the stopping condition for descending down a limb while building more
 *     and more nodes along the way.  
 *     [Default  UniformLeafModel]
 */
DefaultOctreeTemplate
class Octree {
    template <class OctreeType>
    friend void (CMAndVolMeasurement::measure)(OctreeType &, typename OctreeType::OParticle *[]);
  public:
    /* some typedefs for use later. */
    /** Particle type for this Octree. */
    typedef _Particle OParticle;
    /** Extra fields for this Octree. */
    typedef _extraFields OextraFields;
    /** Geometric division definition for this Octree. */
    typedef CMAndVolMeasurement OCMAndVolMeasurement;
    /** Leaf model for this Octree. */
    typedef LeafModel OLeafModel;

    /** we redef the new and delete operators so that we can cache these for
     * multiple tree building.
     */
    REDEF_NEW_DEL_H(Octree);

  public:
    /** Creates a octree node at specfied depth with specified range.
    */
    Octree(int l, int h, int d);

    /** Octree clean up.
     * Frees memory allocated for an octree.
     */
    ~Octree();

    /** build Octree using a quick sort algorithm.
     * @param a
     *    An array of (valid, unique) Particle pointers to be sorted.
     * @param leafModel
     *    The model for determining the rough number of particles per leaf
     *    node [Default UniformLeafModel].
     * 
     * @return int number of leafs in the tree.
     *
     * @see UniformLeafModel
     */
    int buildOctree( _Particle *a[],
                     const LeafModel & leafModel = LeafModel() );

    /** Get number of particles in this octree node.
     * @return Number of particles.
     */
    inline int numberParticles() const {return high-low + 1;}

    /** Write out a GnuPlot plot representing the octree node.
    */
    void gnuplotVol(std::ostream & output, const bool & std_in = true);

    /** Recursively call gnuplotVol for all leaves of this octree.
     * @return true if max_depth < depth of the deepest leaves.
    */
    bool gnuplotLeavesVol(std::ostream & output,
                          const bool & std_in = true,
                          const int & max_depth = -1);

    /** Format the arguments of an splot (GnuPlot) command for plotting boxes
     * representing leaf-volumes.
     * */
    static std::string & createGnuplotLeaveSPLOT(
            std::string & str,
            const std::string & plotopt,
            const int leaf_count,
            const std::string & tree_dat = "");

    /** Overloaded stream output for an Octree.
     * This is a recursive function, plotting out information concerning every
     * node in the tree.
     * */
    friend std::ostream & operator<< <>(std::ostream & output,const Octree & o);

    /** Calculate and return the total enclosed volume of this node.
     * @return Volume enclosed by octree node.
     */
    inline double getVolume() const {
        switch (dimensions) {
            case 3: /* 3d */
                return 1.33333333333333 * M_PI * CUBE(vol_radius);
            case 2: /* 2d */
                return M_PI * SQR(vol_radius);
            case 1: /* 1d */
                return vol_radius;
            default: /* 0 d??? */
                return 0;
        }
    }

    /** Get center of mass location. */
    inline const Vector<double,3> & getRcm() const {return Rcm;}

    /** Get bounded volume. */
    inline const vol_bounds & getVolB() const {return volb;}

    /** Finds the nearest leaf in the tree to specified point.
     * Warning!!!  This must be called after having called findRcmAndVol
     * recursively!
     * @return Octree node nearest to the given point in space.
     */
    Octree & findNearestLeaf(double x, double y, double z);

    /** Finds the nearest leaf in the tree to specified octree node.
     * Warning!!!  This must be called after having called findRcmAndVol
     * recursively! 
     * (And the specified octree node must also have its Rcm calculated.)
     * @return Octree node nearest to the given octree node.
     */
    template <class OTreeType>
    inline Octree & findNearestLeaf(const OTreeType & o) {
        const Vector<double,3> & pp = o.getRcm();
	return findNearestLeaf(pp[X],pp[Y],pp[Z]);
    }

    /** Mark the octree node.
     * A marking currently only changes the way that the node is printed out.
     * In other words, when a node is marked, the gnuplot printout of the node
     * has an X through it diagonally.
     */
    inline void setMarked(bool m) {marked = m;}

    /**
     * findRcmAndVol.
     * measure the center of mass and discover the minimum bounding box for the
     * particles held by this Octree.
     */
    void findRcmAndVol( _Particle *a[], bool recursive = false );

    /** Copy the center of mass position and volume definition from another node.*/
    template <class OTreeType>
    inline void copyRcmAndVol( const OTreeType & o ) {
	memcpy(&volb, &o.getVolB(), sizeof(vol_bounds));
	memcpy(&Rcm, &o.getRcm(), sizeof(Vector<double,3>));
        vol_radius = o.getVolRadius();
        dimensions = o.getDimensions();
	findRcmFinished = true;
    }

    /** The number of dimensions this node lies in. */
    inline const int & getDimensions() const { return dimensions; }

    /** Copy the center of mass position to a particles position.*/
    inline void copyRcm(double  p[]) const {
	memcpy(p, Rcm.val, sizeof(Rcm.val));
    }

    /** Copy the center of mass position to a Vector.*/
    inline void copyRcm(Vector<double,3> &pp) const { pp = Rcm; }

    /** Copy the volume definition.*/
    inline void copyVol(vol_bounds  &vb) const {
	memcpy(&vb, &volb, sizeof(vol_bounds));
    }

    /** Compute the distance between the Octree node and a point. */
    inline double distance(const double x[]) const {
        double dx = (Rcm[X] - x[X]),
               dy = (Rcm[Y] - x[Y]),
               dz = (Rcm[Z] - x[Z]);
        return sqrt( SQR(dx) + SQR(dy) + SQR(dz) );
    }

    /** Compute the distance between a particle and an Octree Rcm. */
    inline double distance(const _Particle & a) {
        double dx = (a.pos[X] - Rcm[X]),
               dy = (a.pos[Y] - Rcm[Y]),
               dz = (a.pos[Z] - Rcm[Z]);
        return sqrt( SQR(dx) + SQR(dy) + SQR(dz) );
    }

    /** Volume effective radius. */
    inline const double & getVolRadius() const { return vol_radius; }

    /** Artificially set the volume dimensions and radius.
     * Note that this should ONLY BE DONE OR VERY RARE OCCASIONS.
     */
    inline void setVolume(const double & r, const int & dim) {
        if (dim < 1 || dim > 3) {
            THROW(std::runtime_error,FILE_LINE "dimensions must be 1, 2, or 3");
        }

        dimensions = dim;
        vol_radius = r;
    }

  public:
    /** Maximum depth of octree.
     * This is mostly used as a debugging flag.  The default value is very
     * large.
     */
    static int maximum_depth;

    /** Balance Criterion.
     * The required ratio of one dimension to the other two in order to split
     * the said dimension to children Octree nodes.
     * Without this criterion, or as the balance_criterion approaches
     * infinity, the resulting tree will have leafnodes with similar
     * dimensions (ratios) to the root node of the tree.
     *
     * The default value is 1.5.
     */
    static double balance_criterion;

    /** Default dimensions.
     * If for some reason the root node contains a zero-D distribution (of one
     * or more particles), this dimensionality of the system will be used.  A
     * warning will be issued (via log_warning) if this does occur.
     */
    static int default_dimensions;

    /** The extrema indices of the particles in this octree node. */
    int low,high;

    /** The depth of this node in an octree. */
    int depth;

    /** The number of leaves of this octree. */
    int leaves;

    /** A pointer to the parent of this node.
     * This is used for traversing the octree in random directions.
     */
    class Octree * parent;

    /** Array of children of this octree.
     * A child pointer might be null, indicating that it doesn't exist.
     */
    class Octree * child[8];

    /** Indicates that this octree node has children. 
     * Warning:  don't change this field explicitly.  Allow for various class
     * routines to do this.  If you don't play nice, then I'll make this field
     * private and only allow access to this via a function call.
     * */
    bool thisHasChildren;

  private:
    bool findRcmFinished;
    Vector<double,3> Rcm;
    vol_bounds volb;
    double vol_radius;
    int dimensions;
    bool marked;
    int sibling_index; /* set by parent during parent->buildOctree(...) */

  public:
    _extraFields extraFields;
};

}/* namespace octree */

#include "Octree.C"

#endif // OCTREE_H
