// -*- c++ -*-
// $Id: Particle.h,v 1.3 2005/06/07 10:00:14 olsonse Exp $
/*
 * Copyright 2000-2004 Spencer Olson
 *
 * $Log: Particle.h,v $
 * Revision 1.3  2005/06/07 10:00:14  olsonse
 * Many changes over the course of time
 *
 * Revision 1.2  2005/01/19 11:44:36  olsonse
 * Added Argon data.
 * Added lines to compile on SunOS.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#ifndef PARTICLE_H
#define PARTICLE_H

#  include "common.h"

#  include <math.h>
#  include <ostream>
#  include <fstream>
#  include <stdexcept>
#  include <cfloat>
#  include <olson-tools/memory.h>
#  include <olson-tools/power.h>
#  include <olson-tools/random/random.h>
#  include <olson-tools/Vector.h>
#  include <olson-tools/ompexcept.h>
#  include <olson-tools/readData.h>

/** Node in a linear database of particle properties.
 * There is a small linear database of particle properties defined in
 * particle.c.
 * TODO:  Make these more unitless, or more natural units.
 * @see Particle::type, getParticleTypeIndex, and particle_info[].
 */
typedef struct {
    /** Name of particle type. */
    char * name;

    /** Integer multiple of particles a this type actually represents. */
    double size;

    /** Mass of particle type.
     * Units:  kg. */
    double mass;

    /** Diameter of particle type.
     * Units:  m.*/
    double diameter;

    /** SWAVE cross section of single particle of this type.
     * Units:  m^2.*/
    double cross_section;

    /** Reference temperature for diameter/cross-section.
     * This is used for the VHS (variable hard sphere model).
     * Please don't forget to give sources :-).
     */
    double T_ref;

    /** Viscosity-temperature power law.
     * @see Graham Bird's \underline{Molecular Gas Dynamics and the Direct
     * Simulation of Gas Flows.}
     */
    double visc_T_law;

    /** The recipricol of Gamma function of (5/2 - viscosity-temperature power law).
     */
    double gamma_visc_inv;

    /** The recipricol of the VSS parameter. */
    double vss_param_inv;

} particle_type_info_t;

/** A linear database of particle properties.
 * @see Particle::type and getParticleTypeIndex.
 */
extern "C" particle_type_info_t particle_info[];

/** Get the index of the particle type with the specified name.
 * @return Index of particle type or -1 if not found.
 * @see Particle::type and particle_type_info_t.
 */
extern "C" int getParticleTypeIndex(const char * name);


/** Just some class to represent particles to sort. */
class ParticleBase {
  public:
    /** Position array.
     * To be clear and consistent, index this array by using the X,Y,Z macros
     * defined in common.h.
     */
    double pos[3];

    /** Velocity array.
     * To be clear and consistent, index this array by using the X,Y,Z macros
     * defined in common.h.
     */
    double vel[3];

    /** note that each module will need to initialize this to an appropriate
     * value.
     * */
    double dt;

    /** Index in the particle_info database for this particle type.
     * @see particle_type_info_t and getParticleTypeIndex.
     */
    int type;

  public:

    /** precscribed to be useable in OctreeMonitor. */
    inline const int & particleType() const {return type;}

    /** Quick way to return physical particle mass. */
    inline const double & mass() const { return particle_info[type].mass; }

    /** Quick way to return physical particle diameter. */
    inline const double & diameter() const { return particle_info[type].diameter; }

    /** Quick way to return physical particle cross-section. */
    inline const double & cross_section() const { return particle_info[type].cross_section; }

    /** Quick way to return physical particle T_ref. */
    inline const double & T_ref() const { return particle_info[type].T_ref; }

    /** Quick way to return physical particle visc_T_law. */
    inline const double & visc_T_law() const { return particle_info[type].visc_T_law; }

    /** Quick way to return physical particle gamma_visc_inv. */
    inline const double & gamma_visc_inv() const { return particle_info[type].gamma_visc_inv; }

    /** Quick way to return physical particle vss_param_inv. */
    inline const double & vss_param_inv() const { return particle_info[type].vss_param_inv; }

    /** Quick way to return representative particle size. */
    inline const double & size() const { return particle_info[type].size; }

    /** Quick way to return effective radius.  Required by Octree. */
    inline double effective_radius() const {
        return 0.5 * particle_info[type].size * particle_info[type].diameter;
    }

    inline ParticleBase(const int & typ = 0, const double & dt_init = 0):
        dt(dt_init), type(typ) {
        memset(pos,0,sizeof(pos));
        memset(vel,0,sizeof(vel));
    }

    /** Copy contructor. */
    inline ParticleBase(const ParticleBase & that) { *this = that; }

    /** Copy operator. */
    inline const ParticleBase & operator=(const ParticleBase & that) {
        memcpy(this,&that,sizeof(ParticleBase));
        return *this;
    }

    inline void randomize(const double & bx = 100,
                          const double & by = 100,
                          const double & bz = 100,
                          const double & bvx = 100,
                          const double & bvy = 100,
                          const double & bvz = 100) {
        /* random values */
        pos[X] = bx*double(rand())/double(RAND_MAX);
        pos[Y] = bx*double(rand())/double(RAND_MAX);
        pos[Z] = bx*double(rand())/double(RAND_MAX);

        vel[X] = bvx*double(rand())/double(RAND_MAX);
        vel[Y] = bvx*double(rand())/double(RAND_MAX);
        vel[Z] = bvx*double(rand())/double(RAND_MAX);
    }

    /** Compute the distance between a Particle and a point. */
    inline double distance(const double x[]) {
        double dx = (pos[X] - x[X]),
               dy = (pos[Y] - x[Y]),
               dz = (pos[Z] - x[Z]);
        return sqrt( SQR(dx) + SQR(dy) + SQR(dz) );
    }

    /** Compute the distance between two Particle\s. */
    static inline double distance(const ParticleBase & a, const ParticleBase & b) {
        double dx = (a.pos[X] - b.pos[X]),
               dy = (a.pos[Y] - b.pos[Y]),
               dz = (a.pos[Z] - b.pos[Z]);
        return sqrt( SQR(dx) + SQR(dy) + SQR(dz) );
    }

    /** Print the three velocity values out seperated by spaces.
    */
    std::ostream & printVelocity(std::ostream & output);
}; // class Particle

/** Print the six position and velocity values out seperated by spaces.
 */
std::ostream & operator<<(std::ostream & output,const ParticleBase & q);

/** Read in the six position and velocity values into ParticleBase class.
 */
std::istream & operator>>(std::istream & input, ParticleBase & p);

/** Load a bunch of particles from a flat file.  The returned value is a
 * pointer to a linked list of particles.
 */
template <class _Particle>
_Particle * loadParticles(const char * filename,
                          const int & type = 0,
                          const double & dt = 0,
                          int * Nout = NULL) {
    _Particle * retval = NULL;

    std::ifstream input(filename);

    retval = readData<_Particle>(input, Nout);

    /* set the type and dt */
    _Particle * ptr = retval;
    while (ptr) {
        ptr->type = type;
        ptr->dt   = dt;
        ptr = ptr->next;
    }

    return retval;
}



/** Split a particle into sub-particles.
 * To make things easy, we will split particles into square volume
 * distributions.  I don't want to have to invert any spherical
 * distribution.  The template parameter allows for splitting various types of
 * particles with the same code.
 *
 * @param old_size
 *     The old size of the particles.
 *
 * @param new_size
 *     The new size of the particles.
 *
 * @param density
 *     Specify each of the linear densities to use when creating the new
 *     particles.  A linear density of DBL_MAX is necessary to elimintate
 *     a distribution in a specific dimension.
 *
 * @param seed
 *     The seed Particle to split.
 *
 * @param count_seed
 *     Whether the seed Particle should also be in the count of total
 *     split particles [Default true].
 *
 * @returns
 *     A linked list of the newly created particles.
 */
template <class _Particle>
inline _Particle * split(const double & old_size,
                               const double & new_size,
                               const Vector<double,3> & density,
                               const _Particle & seed,
                               const bool count_seed = true) {
    /* FIXME:  perhaps we should do a poisson distrib around the new
     * number instead of using rint(). */
    int N = (int)rint(old_size/new_size);

    if (N < 1) {
        THROW(std::runtime_error,"We can only split particles, not join them!");
    }

    if (count_seed) N--;

    Vector<double,3> dx(0.5);
    dx.compDiv(density);
    /* I do this case just to make sure that it is set to exactly zero. */
    int dim = 0;
    for (int i = 0; i < 3; i++) {
        if (density[i] == DBL_MAX) {
            dx[i] = 0;
        } else dim++;
    }
    dx *= fast_pow(old_size,1.0/double(dim));

    _Particle * list = NULL;
    Vector<double,3> & x0 = V3C(seed.pos);
    Vector<double,3> & v0 = V3C(seed.vel);
    for (int i = 0; i < N; i++) {
        _Particle * p = new _Particle(seed);
        p->next = list;
        list = p;

        Vector<double,3> & x  = V3C(p->pos);
        Vector<double,3> & v  = V3C(p->vel);

        x[X] = MTRNGrand() * 2.0 * dx[X];
        x[Y] = MTRNGrand() * 2.0 * dx[Y];
        x[Z] = MTRNGrand() * 2.0 * dx[Z];
        x -= dx;
        x += x0;
        v = v0;
    }

    return list;
}

#  if defined(USE_MPI)

#include "ParticleNode.h" 

/** Particle size specialization for ParticleNode (MPI) use. */
template<>
inline int particle_packed_size<ParticleBase>(MPI::Comm & comm) {
    return MPI::DOUBLE.Pack_size(7,comm) + MPI::INT.Pack_size(1,comm);
}

/** Particle unpack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_unpack<ParticleBase>(const uint8_t * buffer,
                                      const int & bufsz,
                                      ParticleBase * p,
                                      int & bpos,
                                      MPI::Comm & comm ) {
    /* unpack pos[3], vel[3], and dt */
    MPI::DOUBLE.Unpack(buffer, bufsz, p->pos, 7, bpos, comm);
    MPI::INT.Unpack(buffer, bufsz, &p->type, 1, bpos, comm);
}

/** Particle pack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_pack<ParticleBase>(uint8_t * buffer,
                                    const int & bufsz,
                                    const ParticleBase * p,
                                    int & bpos,
                                    MPI::Comm & comm) {
    /* pos[3], vel[3], and dt */
    MPI::DOUBLE.Pack(p->pos, 7, buffer, bufsz, bpos, comm);
    /* type */
    MPI::INT.Pack(&p->type, 1, buffer, bufsz, bpos, comm);
}

#  endif // USE_MPI


/** Just some class to represent particles to sort. */
class Particle : public ParticleBase {
  public:
    /** Pointer to next Particle in a linked list of Particle\s.
     */
    class Particle * next;

    /** we redef the new and delete operators so that we can cache these for
     * multiple tree building.
     * Note:  it is important that this is declared on the latter end of the
     * object so that passing a Particle to code written in non-c++ (eg.
     * fortran) will work.
     */
    REDEF_NEW_DEL_H(Particle);
  public:

    inline Particle(const int & typ = 0, const double & dt_init = 0):
        ParticleBase(typ,dt_init), next(NULL) { }

    /** Copy contructor--does not copy the next field. */
    inline Particle(const Particle & that) : next(NULL) { *this = that; }

    /** Copy operator--does NOT change the next field. */
    inline const Particle & operator=(const Particle & that) {
        register Particle * ptr = next;
        ParticleBase::operator=(that);
        next = ptr;
        return *this;
    }
}; // class Particle

/** Read in the six position and velocity values into a newly created particle.
 * This function satisfies the requirements for the readData routines in the
 * olson-tools package.
 */
std::istream & operator>>(std::istream & input,Particle *& p);

#  if defined(USE_MPI)

/** Particle size specialization for ParticleNode (MPI) use. */
template<>
inline int particle_packed_size<Particle>(MPI::Comm & comm) {
    return particle_packed_size<ParticleBase>(comm);
}

/** Particle unpack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_unpack<Particle>(const uint8_t * buffer,
                                      const int & bufsz,
                                      Particle * p,
                                      int & bpos,
                                      MPI::Comm & comm ) {
    /* unpack pos[3], vel[3], and dt */
    particle_unpack<ParticleBase>(buffer, bufsz, (ParticleBase*)p, bpos, comm);
}

/** Particle pack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_pack<Particle>(uint8_t * buffer,
                                    const int & bufsz,
                                    const Particle * p,
                                    int & bpos,
                                    MPI::Comm & comm) {
    particle_pack<ParticleBase>(buffer, bufsz, (ParticleBase*)p, bpos, comm);
}

#  endif // USE_MPI


#endif // PARTICLE_H
