// -*- c++ -*-
// $Id$
/*
 * Copyright 2005 Spencer Olson
 *
 * $Log$
 *
 */

#ifndef PARTICLENODE_H
#define PARTICLENODE_H

#include <math.h> 
#include <mpi.h> 
#include <string.h>
#include <unistd.h>
#include <sys/times.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <signal.h>

#include <iostream>
#include <vector>
#include <stdexcept>
#include <cfloat>


#include <olson-tools/Vector.h>
#include <olson-tools/strutil.h>
#include <olson-tools/power.h>
#include <olson-tools/random/random.h>
#include <olson-tools/logger.h>
#include <olson-tools/ompexcept.h>
#include <olson-tools/Distribution.h>

#include "common.h"
#include "BoundaryObject.h"

#include "debug_util.h"

/** \example mpi4dsmc.C
 * Very simple demonstration simulation that uses MPI.  This simulation also
 * prints out information for each node to demonstrate the function of the MPI
 * load balancing code. 
 */

static const double _for_mpi_seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);

/** A simple Boundary class for MPI domains.
 * This class represents the domain division between two different MPI nodes.
 */
template <class _Particle>
class MPIBoundary : public octree::BoundaryObject<_Particle> {
  public:
    typedef     octree::BoundaryObject<_Particle> super;

    /** default constructor. */
    MPIBoundary() : super(), left_side(false), neighbor(-1), location(0), physical_orientation(X) {
        super::keepRemoved = true;
    }

    /** Initialize the MPI boundary. This function is routinely called and
     * re-called from the decomposeEntireDomain(...) function as boundaries
     * are reassigned.
     */
    inline void initBoundary(const bool & left, const int & phys_or, const double & loc) {
        left_side = left;
        physical_orientation = phys_or;
        location = loc;

        if (left_side)  {
            super::volb_orig.min[physical_orientation] = -DBL_MAX;
            super::volb_orig.max[physical_orientation] = location;
        } else {
            super::volb_orig.min[physical_orientation] = location;
            super::volb_orig.max[physical_orientation] = +DBL_MAX;
        }

        super::volb_orig.max[(physical_orientation+1)%3] = +DBL_MAX;
        super::volb_orig.max[(physical_orientation+2)%3] = +DBL_MAX;

        super::volb_orig.min[(physical_orientation+1)%3] = -DBL_MAX;
        super::volb_orig.min[(physical_orientation+2)%3] = -DBL_MAX;
    }

    /** destructor. */
    virtual ~MPIBoundary()  { }

    /** interaction mechanism. */
    virtual enum octree::INTERACTION interact(_Particle & pold, _Particle & p) {
        if (left_side && p.pos[physical_orientation] < location) {
            /* the Particle crossed the boundary, so remove it. */
            return octree::REMOVED;
        } else if ( !left_side && p.pos[physical_orientation] > location) {
            /* the Particle crossed the boundary, so remove it. */
            return octree::REMOVED;
        } else return octree::UNCHANGED;
    }

    /** left or right side of node? */
    bool left_side;

    /** rank of neighbor through this boundary. */
    int neighbor;

    /** The location of the boundary. */
    double location;

    /** physical_orientation. */
    int physical_orientation;
};


/** Generic Particle is assumed to have position[3] and velocity[3].
 * For non-generic Particle/s, provide a template specialization.
 * Almost valid.
 */
template<class _Particle>
inline int particle_packed_size(MPI::Comm & comm) {
    _Particle::implement_your_own_method_for_MPI = 0;
    return MPI::DOUBLE.Pack_size(6,comm);
}

/** Generic Particle is assumed to have position[3] and velocity[3].
 * For non-generic Particle/s, provide a template specialization.
 */
template<class _Particle>
inline void particle_unpack(const uint8_t * buffer,
                            const int & bufsz,
                            _Particle * p,
                            int & bpos,
                            MPI::Comm & comm ) {
    /* unpack pos[3], vel[3] */
    _Particle::implement_your_own_method_for_MPI = 0;
    MPI::DOUBLE.Unpack(buffer, bufsz, p->pos, 6, bpos, comm);
}

/** Generic Particle is assumed to have position[3] and velocity[3].
 * For non-generic Particle/s, provide a template specialization.
 */
template<class _Particle>
inline void particle_pack(uint8_t * buffer,
                          const int & bufsz,
                          const _Particle * p,
                          int & bpos,
                          MPI::Comm & comm) {
    /* pos[3], vel[3] */
    _Particle::implement_your_own_method_for_MPI = 0;
    MPI::DOUBLE.Pack(p->pos, 6, buffer, bufsz, bpos, comm);
}


/* ************* CORE DUMPING IN MPI/PBS ***************** */
/* This core dumping logic is supposed to make it easier to obtain a random
 * core dump when mpi code is running  under the influence of PBS management.
 * According to the qdel man page, qdel sends SIGTERM first and then SIGKILL
 * after a specified delay.  We provide the core dump by catching the SIGTERM
 * and subsequently calling abort() which causes the core dump. 
 * */
static pthread_t main_tid = 0;

/** Calculate pi.  This is a hack just to slow the signal handler down a bit.
 * We can't do sleep because core doesn't get dumped if we do.  The number of
 * steps will have to be tuned to the particular CPU that is being used.
 * On kamala, (num_steps = 1000000000) takes approximately 7 seconds.  This
 * should allow all processes on all machines to recieve the TERM signal and
 * react appropriately. */
inline double calculate_pi () {
    const int num_steps = 100000000;
    double sum = 0.0;
    double step = 1.0/(double) num_steps;
    for (int i=0;i< num_steps; i++){
        double x = (i+0.5)*step;
        sum = sum + 4.0/(1.0+x*x);
        }
    return step * sum;
}

/** Cause the current process to dump core (if ulimit is set correctly). */
void dumpCore(int inSignal) {
    if (main_tid == pthread_self()) {
        //log_severe("calculating pi to slow us down a bit");
        //log_severe("pi is %g, now isn't that special", calculate_pi());
        log_severe("dumping core now");
        abort();
    }
}

/** Prepare the thread to dump core in a PBS queuing system.  I could only get
 * this process to work on the head node. Oh, well.  It was still helpful for
 * debugging. 
 */
void prepareCoreDumps(const pthread_t & tid) {
    main_tid = tid;
    /* dumping core on SIGTERM allows us to dump core in a PBS system. */
    signal(SIGTERM, dumpCore);

    struct rlimit rlim = {0,0};
#if 0
    getrlimit(RLIMIT_CORE, &rlim);
    rlim.rlim_cur = rlim.rlim_max;
    setrlimit(RLIMIT_CORE, &rlim);

#endif
    getrlimit(RLIMIT_CORE, &rlim);

    log_info("core dumping limits (current/max) : (%lu/%lu)", rlim.rlim_cur, rlim.rlim_max);
}


/** A class for compartmentalizing the parallel code for a 'Particular'
 * dynamics simulation.
 */
template <class _Particle, class ParticleContainer>
class ParticleNode {
  public:
    /** Default constructor defaults to  1-D automatic load balancing scheme.
     * All available processors will be used if possible.
     */
    ParticleNode() { class_init(); }

    /** Destructor. */
    ~ParticleNode() {
        if (mpi_info.proc_map != NULL) {
            delete[] mpi_info.proc_map;
            mpi_info.proc_map = NULL;
        }
        /* We are going to perform a sync on the event of the exit so that ALL
         * processes have the chance to finish business before the MPI
         * overlord kills siblings. */
        MPI::COMM_WORLD.Barrier();

        if (initialized) {
            MPI::Finalize(); 
        }
    }

  private:
    inline void class_init() {
        initialized             = false;
        cpu_time                = 0;
        physical_orientation    = X;
        load_balance_rshift      = (+DBL_MAX);
        load_balance_lshift      = (+DBL_MAX);
        n_active_proc           = 0;
        load_shift_trigger_frac = 0.1;
        tot_shift_err           = 0.0;
        stddev2_tot_shift_err    = 0.0;
        shift_err_hfactor       = 0.50;
        amin_N_per_proc         = 10000.0;
        max_packet_length       = 5000; /* roughly 500kB */
        setNullVolume();

        /* init mpi_info */
        mpi_info.Np = 0;
        mpi_info.arank = -1;
        mpi_info.rank = -1;
        mpi_info.aroot = -1;
        mpi_info.proc_map = NULL;

        /* working directory changes. */
        wdir_change = true;
        wdir_perm = S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

        /* We default to core dumping when sigterm is received. */
        sigterm_core_dumps = true;
    }

    inline void setNullVolume() {
        /* I'm just hoping that this will ensure that a test for
         * volb.contains(...) always fails.
         */
        for (int i = X; i <= Z; i++) volb.min[i] = +DBL_MAX;
        for (int i = X; i <= Z; i++) volb.max[i] = -DBL_MAX;
    }

    inline void setInfVolume() {
        /* I'm just hoping that this will ensure that a test for
         * volb.contains(...) always fails.
         */
        for (int i = X; i <= Z; i++) volb.min[i] = -DBL_MAX;
        for (int i = X; i <= Z; i++) volb.max[i] = +DBL_MAX;
    }

    inline void setupActiveComm(const int & n_proc) {
        /* We'll free the previous communicator first off. This proc may not
         * be active anymore.  Note that although the comm seemed to be
         * created in all processes regardless of their inclusion in the
         * group, trial and error (by me) has shown that we should only free
         * the comm from processes that were included in the comm.
         */
        if (mpi_info.arank >= 0) mpi_info.comm.Free();

        int * granks = new int[n_proc];
        for (int i = 0; i < n_proc; i++) {
            granks[i] = i;
        }

        log_finest(FILE_LINE ":  MPI::Group g = MPI::COMM_WORLD.Get_group().Incl(n_proc, granks);");
        MPI::Group g = MPI::COMM_WORLD.Get_group().Incl(n_proc, granks);

        log_finest(FILE_LINE ":  mpi_info.comm   = MPI::COMM_WORLD.Create(g);");
        mpi_info.comm   = MPI::COMM_WORLD.Create(g);

        if (mpi_info.rank >= n_proc) {
            mpi_info.arank = -1;
        } else {
            mpi_info.arank  = mpi_info.comm.Get_rank();
        }

        delete[] granks;

        /* make sure that all nodes, even inactive ones, know who the root of
         * the active nodes is. This communication is rooted from main root. */
        {
            int amIaroot = (mpi_info.arank == 0 ? mpi_info.rank : -1);
            log_finest(FILE_LINE ":  MPI::COMM_WORLD.Reduce(&amIaroot, &mpi_info.aroot, 1, MPI::INT, MPI::MAX, 0);");
            MPI::COMM_WORLD.Reduce(&amIaroot, &mpi_info.aroot, 1, MPI::INT, MPI::MAX, 0);
            log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&mpi_info.aroot, 1, MPI::INT, 0);");
            MPI::COMM_WORLD.Bcast(&mpi_info.aroot, 1, MPI::INT, 0);
        }

        if (mpi_info.proc_map != NULL) {
            delete[] mpi_info.proc_map;
        }

        mpi_info.proc_map = new int[n_proc];

        if (mpi_info.arank == 0) {
            mpi_info.proc_map[0] = mpi_info.rank;
            for (int i = 1; i < n_proc; i++) {
                mpi_info.comm.Recv( (mpi_info.proc_map + i), 1, MPI::INT, i, msgtags::PROC_MAP );
            }
        } else if (mpi_info.arank > 0) {
            mpi_info.comm.Send( &mpi_info.rank, 1, MPI::INT, 0, msgtags::PROC_MAP );
        }

        /* broadcast the arank to rank mappings to everyone. This is used for
         * decompose. */
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(mpi_info.proc_map, n_proc, MPI::INT, mpi_info.aroot);");
        MPI::COMM_WORLD.Bcast(mpi_info.proc_map, n_proc, MPI::INT, mpi_info.aroot);
    }


  public:

    /** The initialization function for the MPI node must be called before
     * calculations are done. */
    void init(int & argc, char **& argv) {
        MPI::Init(argc, argv); 
        log_finest(FILE_LINE ":  mpi_info.Np   = MPI::COMM_WORLD.Get_size();");
        mpi_info.Np   = MPI::COMM_WORLD.Get_size(); 
        log_finest(FILE_LINE ":  mpi_info.Np   = MPI::COMM_WORLD.Get_rank();");
        mpi_info.rank = MPI::COMM_WORLD.Get_rank(); 

        initialized = true;

        /* default to only one active processor. */
        n_active_proc = 1;
        if (mpi_info.rank == 0) {
            setInfVolume();
        } else {
            setNullVolume();
        }

        /* create and change to working directory for this node */
        if (wdir_change) {
            mkdir(to_string(mpi_info.rank).c_str(), wdir_perm);
            chdir(to_string(mpi_info.rank).c_str());
        }

        setupActiveComm(n_active_proc);

        if (sigterm_core_dumps) {
            prepareCoreDumps(pthread_self());
        }
    }

  private:
    /** the various message tags that will be used here. */
    class msgtags {
      public:
        enum tag {
            EXCHANGE_LENGTH,
            EXCHANGE_TIME,
            EXCHANGE_NPART,
            EXCHANGE_MTRANSIT,
            EXCHANGE_NODETYPE,
            SHIFT,
            PARTICLES,
            DECOMPOSE_TEST,
            BOUNDARY_MIN,
            BOUNDARY_MAX,
            MAX_TRANSIT,
            PROC_MAP
        };
    };

    class ShuffleQ {
      public:
        inline ShuffleQ(): neighbor(0), q(NULL) {}
        int neighbor;
        _Particle * q;
    };

    enum NODE_TYPE {
        LEFT_NODE,  /* A node lowest in position space. */
        MID_NODE,   /* A node somewhere between others. */
        RIGHT_NODE, /* A node highest in position space. */
        LONELY_NODE /* A node all by itsself.  How sad. */
    };

    inline NODE_TYPE node_type() const {
        bool left  = mpi_boundaries[0].neighbor < 0;
        bool right = mpi_boundaries[1].neighbor < 0;

        if (!left && !right)    return MID_NODE;
        else if (!left)         return RIGHT_NODE;
        else if (!right)        return LEFT_NODE;
        else                    return LONELY_NODE;
    }

    /* The length of the nodes computational domain.  This function fixes the
     * result for edge nodes by calling the containers .getVolB function to
     * return the actual extent of the volume of particles in the node.
     *
     * WARNING:  This function returns zero length for edge nodes with no
     * particles (and the half width of a particle if only one particle).
     */
    inline double node_length() {
        double x_i = -DBL_MAX, x_f = +DBL_MAX;
        get_node_edges(x_i, x_f);
        return x_f - x_i;
    }

    /* The edge positions of the node's computational domain.  This function fixes the
     * result for edge nodes by calling the containers .getVolB function to
     * return the actual extent of the volume of particles in the node.
     *
     * WARNING:  This function returns zero length for edge nodes with no
     * particles (and the half width of a particle if only one particle).
     *
     * WARNING:  This function erases the current octree beyond the root node
     * for all (both) edge nodes.
     */
    inline void get_node_edges(double & x_i, double & x_f) {
        x_i = volb.min[physical_orientation];
        x_f = volb.max[physical_orientation];

        switch(node_type()) {
            case LEFT_NODE:
                /* I'm on the far left, so get my min from the container. */
                switch (container.numberParticles()) {
                    case 0:
                        x_i = x_f;
                        break;
                    case 1: {
                        _Particle & p = *(container.getParticles()[0]);
                        x_i = std::min(x_f, p.pos[physical_orientation] - p.effective_radius());
                        break;
                    }
                    default:
                        container.refreshCurrentRcmAndVol();
                        /* FIXME:  Why do we need the std::max()?  */
                        x_i = std::min(x_f, container.getVolB().Min(physical_orientation));
                } /* switch */
                break;

            case RIGHT_NODE:
                /* I'm on the far right, so get my max from the container. */
                switch (container.numberParticles()) {
                    case 0:
                        x_f = x_i;
                        break;
                    case 1: {
                        _Particle & p = *(container.getParticles()[0]);
                        x_f = std::max(x_i, p.pos[physical_orientation] + p.effective_radius());
                        break;
                    }
                    default:
                        container.refreshCurrentRcmAndVol();
                        /* FIXME:  Why do we need the std::max()?  */
                        x_f = std::max(x_i, container.getVolB().Max(physical_orientation));
                } /* switch */
                break;

            case LONELY_NODE:
                /* I'm the only node out there.  Where is everybody?  */
                switch (container.numberParticles()) {
                    case 0:
                        x_f = x_i = 0;
                        break;
                    case 1: {
                        _Particle & p = *(container.getParticles()[0]);
                        x_i = p.pos[physical_orientation] - p.effective_radius();
                        x_f = p.pos[physical_orientation] + p.effective_radius();
                        break;
                    }
                    default:
                        container.refreshCurrentRcmAndVol();
                        /* FIXME:  Why do we need the std::max()?  */
                        x_i = container.getVolB().Min(physical_orientation);
                        x_f = container.getVolB().Max(physical_orientation);
                } /* switch */
                break;

            default:
            case MID_NODE:
                break;
        }/* switch */
    }/* get_node_edges */

    /** Measure the balance between two neighbors and attempt to equalize
     * their work.
     *
     * The primary feedback mechanism has been the system time required for
     * the previous evolve iteration as returned by the system call 'times'.
     *
     * It seems that the particle number should also play a role.  For young
     * simulations (for few particles), the cpu_time fluctuates a lot.
     *
     * TODO:  uncomment the code below which attempts to take in account for
     * the particle/node number.  This code will have to be slightly fixed up
     * and then heavily tested to see if it can calm down the startup period
     * of a simulation.  This might not be necessary with the better
     * decomposition code (that might work).
     */
    double shiftNeighborBoundary(MPIBoundary<_Particle> & bo) {
        /* find out how much time the two nodes took. */
        double hisT = 0;
        MPI::Request request = mpi_info.comm.Irecv( &hisT, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_TIME );
        mpi_info.comm.Send( &cpu_time, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_TIME );
        request.Wait();

#if 0
        /* find out how many particles the two nodes have. */
        double hisN = 0;
        static double myN = 0;
        myN *= (1.0 - PID.i_N);
        myN += PID.i_N*container.numberParticles();

        request = mpi_info.comm.Irecv( &hisN, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_NPART );
        mpi_info.comm.Send( &myN, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_NPART );
        request.Wait();
#endif

        /* find out how small the two nodes can become:
         * We are only worried about the case when our node becomes so small
         * that a Particle from the neighbor flys through us in one time step.
         * Thus, our minimum size is his_max_transit.
         *
         * We need to ensure that max_transit is a positive quantity.
         */
        double my_max_transit = fabs(container.max_particle_transit(physical_orientation)),
               his_max_transit = 0;
        request = mpi_info.comm.Irecv( &his_max_transit, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_MTRANSIT );
        mpi_info.comm.Send( &my_max_transit, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_MTRANSIT );
        request.Wait();

        /* peers must know if the other is an edge node. */
        int his_node_type = MID_NODE,
            my_node_type = node_type();
        request = mpi_info.comm.Irecv( &his_node_type, 1, MPI::INT, bo.neighbor, msgtags::EXCHANGE_NODETYPE );
        mpi_info.comm.Send( &my_node_type, 1, MPI::INT, bo.neighbor, msgtags::EXCHANGE_NODETYPE );
        request.Wait();

        /* find out how big the two nodes are. */
        double my_dx = node_length(), his_dx = 0;

        if (my_dx < 0) {
            THROW(std::runtime_error,FILE_LINE " my length < 0!");
        }

        request = mpi_info.comm.Irecv( &his_dx, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_LENGTH );
        mpi_info.comm.Send( &my_dx, 1, MPI::DOUBLE, bo.neighbor, msgtags::EXCHANGE_LENGTH );
        request.Wait();

        double L = my_dx + his_dx;
        double retval = 0.0;

        /* adjust boundaries to provide load balancing. */
        {
            /* note that for load_balance_[rl]shift > 0, the current node
             * decreases in size. */
            register double t = cpu_time + hisT,
                           dt = cpu_time - hisT;
#if 0
                           NT = myN + hisN,
                           dN = myN - hisN;
            double load_balance_shift
                    = (1.0 - PID.d) * PID.p_T * dt * (L/(t + 1e-3))
                    +        PID.d  * PID.p_N * dN * (L / (NT + 1e-3));
#endif

            PID & pid = bo.left_side ? lPID : rPID;

#if 0
            double load_balance_shift
                    = std::min(pid.Kp, 0.251665*fast_pow(t,0.327544))
                    * 0.5 * (L/(t + 1e-3))
                    * (dt + pid.IntegralError(dt)/pid.Ti - pid.Td*pid.dE(dt));
#else

            //double pterm = std::min(pid.Kp, 0.251665*fast_pow(t,0.327544));
            //double pterm = std::min(pid.Kp, 0.25*fast_pow(t,-0.33333333333333333));
            register double pterm = pid.Kp * dt;
            register double iterm = pid.IntegralError(dt)/pid.Ti;
            register double dterm = pid.Td*pid.dE(dt);
            double load_balance_shift = 0.5 * (L/(t + 1e-3)) * ( pterm + iterm - dterm );
#if 0
            static double int_response = 0.0;
            int_response += load_balance_shift;
            std::cout << getpid()
                      << "0.5 * ("<<L<<"/("<<t<<" + 1e-3)) * (  "<<dt<<" + "<<iterm<<" - "<<dterm<<") * "<<pterm
                      << std::endl;
            std::cout << getpid()
                      << ": t = " << t
                      << ", dt = " << dt
                      << ", L = " << t
                      << ", pterm = " << pterm
                      << ", iterm = " << iterm
                      << ", dterm = " << dterm
                      << ", shift = " << load_balance_shift
                      << std::endl;
#endif
#endif


#if 0
            /* minimize the probabilty of changing.  Only the left node does
             * this.  This uses an analogy to cooling of a physical process.
             * The hotter it is, the more we allow changes.  The probability
             * of the change is from the partition function : exp(-V/kT) */
            if (!bo.left_side) {
                double temperature = dt / (t + 1e-15);
                double V = 2.0*load_balance_shift / (L + 1e-15);
                int do_shift = 1;

                if (MTRNGrand() > exp(-V/temperature)) {
                    load_balance_shift = 0.0;
                    do_shift = 0;
                }

                /* send result of probability test. */
                mpi_info.comm.Send( &do_shift, 1, MPI::INT, bo.neighbor, msgtags::SHIFT );
            } else {
                int do_shift = 0;

                /* receive the result of the probability test. */
                mpi_info.comm.Recv( &do_shift, 1, MPI::INT, bo.neighbor, msgtags::SHIFT );

                if (!do_shift) load_balance_shift = 0.0;
            }
#endif

            if (my_max_transit >= his_dx || his_max_transit >= my_dx) {
                log_severe(FILE_LINE "(%gs):  containers too small.  "
                    "(dx,transit) : my(%g,%g), his(%g,%g)",
                    container.getAbsoluteTime(),
                    my_dx, my_max_transit,
                    his_dx, his_max_transit
                    );
            }

            /* perform simple checks on shift. */
            if (load_balance_shift > 0) {
                if ( my_node_type == MID_NODE && (my_dx - load_balance_shift) < 1.5*his_max_transit ) {
                    /* keep the shift positive. */
                    log_warning(
                        FILE_LINE "(%gs):  shift is considered too large(+).  Reseting PID.",
                        container.getAbsoluteTime());
                    load_balance_shift = std::max(0.0,my_dx - 1.5*his_max_transit);
                    pid.reset();
                }
            } else {
                if ( his_node_type == MID_NODE && (his_dx + load_balance_shift) < 1.5*my_max_transit ) {
                    /* keep the shift negative. */
                    log_warning(
                        FILE_LINE ":  shift is considered too large(-).  Reseting PID.",
                        container.getAbsoluteTime());
                    load_balance_shift = std::min(0.0,-his_dx + 1.5*my_max_transit);
                    pid.reset();
                }
            }

            /* set boundary surface. */
            if (bo.left_side) {
                bo.location += load_balance_shift;
                bo.volb_orig.max[physical_orientation] = bo.location;
                volb.min[physical_orientation] = bo.location;

                load_balance_lshift = load_balance_shift;
            } else {
                bo.location -= load_balance_shift;
                bo.volb_orig.min[physical_orientation] = bo.location;
                volb.max[physical_orientation] = bo.location;

                load_balance_rshift = load_balance_shift;
            }

            retval = fabs(load_balance_shift) / (my_dx + 1e-15);
        }

        return retval;
    }/* shiftNeighborBoundary */

  private:
    class ProcDistributionFunction {
      private:
        class pdf_entry {
          private:
            double x_i;
          public:
            double x_f;
            double value;
            int arank;

            pdf_entry(const double & _x_i,
                      const double & _x_f,
                      const double & cpu_time,
                      const double & max_transit,
                      const int    & _arank ) :
                x_i(_x_i), x_f(_x_f), arank(_arank) {
                /* we have to take care of all things that might be zero
                 * because the Distribution class doesn't like all zeros and
                 * it also doesn't like NANs. */

                double dx = 1e-300 + x_f - x_i;
                /* Np_i is the number of processors that could be supported by
                 * this particular node.  It is clamped here to no more than
                 * one because we only want to modify the pdf when there
                 * should be less CPUs at a given location.  */
                double Np_i = std::min(1.0, dx / (1e-300 + max_transit));
                value = ( 1e-300 + cpu_time )  / (dx * Np_i);
            }

            pdf_entry(const pdf_entry & that) {
                *this = that;
            }

            pdf_entry & operator=(const pdf_entry & that) {
                x_i = that.x_i;
                x_f = that.x_f;
                value = that.value;
                arank = that.arank;
                return *this;
            }
        };

        std::vector<pdf_entry> pdf;

        bool flat_dist;

      public:
        ProcDistributionFunction(const bool & _flat_dist,
                                 const double & _x_i,
                                 const double & _x_f,
                                 const double & _cpu_time,
                                 const double & _max_transit,
                                 MPI::Intracomm & comm) {
            flat_dist = _flat_dist;

            /* add the aroot info. */
            pdf.push_back(pdf_entry(_x_i, _x_f, _cpu_time, _max_transit, 0));

            /* recv and add the rest of the info. */
            for (int arank = 1; arank < comm.Get_size(); arank++) {
                double x_i = 0.0,
                       x_f = 0.0,
                       cpu_time = 0.0,
                       max_transit = 0.0;

                /* need to obtain the next boundary info. */
                comm.Recv( &x_i,          1, MPI::DOUBLE, arank, msgtags::BOUNDARY_MIN );
                comm.Recv( &x_f,          1, MPI::DOUBLE, arank, msgtags::BOUNDARY_MAX );
                comm.Recv( &cpu_time,     1, MPI::DOUBLE, arank, msgtags::EXCHANGE_TIME );
                comm.Recv( &max_transit,  1, MPI::DOUBLE, arank, msgtags::MAX_TRANSIT );
                pdf.push_back(pdf_entry(x_i, x_f, cpu_time, max_transit, arank));
            }
        }

        /** a function that returns the CPU time required per length at a given
         * location in the entire domain.  This is to pass in for the Distribution
         * class which we will use to invert the workload distribution. */
        inline double distrib(const double & x) const {
            if (flat_dist) return 0.5;

            /* find the processor */
            for (unsigned int i = 0; i < pdf.size(); i++) {
                const pdf_entry & pdfi = pdf[i];
                if (x < pdfi.x_f) return pdfi.value;
            }

            return pdf.back().value;
        }
    };

  public:
    /** Request decomposition of the computational domain into subdomains
     * owned by individual processors. 
     *
     * @param max_active_proc
     *     Limit the number of available processors to decompose the domain
     *     into.  This is helpful when the user is initializing the domain
     *     from a known configuration where the number of processors was
     *     limited due to something like non-uniform gas properties.
     */
    void decomposeEntireDomain(int max_active_proc = 0, bool flat_dist = false, bool force_max = false) {
        log_info("Decomposing entire domain at time %g s", container.getAbsoluteTime());

        /* ********  get total number of particles.  ******** */
        double myN = container.numberParticles();

        if (myN > 0 && mpi_info.arank < 0) {
            THROW(std::runtime_error,"I'm not active, but I have some particles!");
        }

        double N = 0;

        /* Note that in the comm.Reduce, '0' means aroot.  Thus, when we
         * disseminate the info to the non-active nodes, we must do it from
         * aroot (relative to the COMM_WORLD context). */
        if ( mpi_info.arank >= 0 ) {
            log_info("decompose starts me with %d particles", (int)myN);
            log_finest(FILE_LINE ":  mpi_info.comm.Reduce(&myN, &N, 1, MPI::DOUBLE, MPI::SUM, 0);");
            mpi_info.comm.Reduce(&myN, &N, 1, MPI::DOUBLE, MPI::SUM, 0);
        }
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&N, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&N, 1, MPI::DOUBLE, mpi_info.aroot);

        /* so, if this node is no longer active, it doesn't need to complete
         * too much.  BUT, it will have to dumps its Particles somewhere.
         * Perhaps it is best that an inactive node completes everything,
         * including the sending and receiving of particles so that it is able
         * to dump its particles. */

        /* ******** Now compute the lengths of nodes of system. ********* */

        double L_min    = +DBL_MAX;     /* min length of nodes. */
        double x_i      = -DBL_MAX;     /* min position in domain. */
        double x_f      = +DBL_MAX;     /* max position in domain. */
        double myx_i    = 0;            /* my min position (arank >=0) */
        double myx_f    = 0;            /* my max position (arank >=0) */
        if ( mpi_info.arank >= 0 ) {
            get_node_edges(myx_i, myx_f);
            double myL = myx_f - myx_i;

            mpi_info.comm.Reduce(&myL, &L_min, 1, MPI::DOUBLE, MPI::MIN, 0);
            mpi_info.comm.Reduce(&myx_i, &x_i, 1, MPI::DOUBLE, MPI::MIN, 0);
            mpi_info.comm.Reduce(&myx_f, &x_f, 1, MPI::DOUBLE, MPI::MAX, 0);
        }

        MPI::COMM_WORLD.Bcast(&L_min, 1, MPI::DOUBLE, mpi_info.aroot);
        MPI::COMM_WORLD.Bcast(&x_i, 1, MPI::DOUBLE, mpi_info.aroot);
        MPI::COMM_WORLD.Bcast(&x_f, 1, MPI::DOUBLE, mpi_info.aroot);

        double L_total = x_f - x_i;

        /* ******** Determine workload distribution ******** */
        Distribution * pDist = NULL;

        if (mpi_info.arank >= 0) {
            double my_max_transit = container.max_particle_transit(physical_orientation);

            if (mpi_info.arank == 0) {
                const double MIN_BIN_PER_NODE = 10.0;
                int nbins = (int) ( MIN_BIN_PER_NODE * ( L_total / L_min ) );

                pDist = new Distribution (
                    ProcDistributionFunction(flat_dist, myx_i, myx_f, cpu_time, my_max_transit, mpi_info.comm),
                    x_i, x_f, nbins
                );
            } else {
                /* We were active, so send our latest boundaries and cpu_time
                 * to aroot.  We will be answered in order of our arank.  */
                mpi_info.comm.Send( &myx_i, 1, MPI::DOUBLE, 0, msgtags::BOUNDARY_MIN );
                mpi_info.comm.Send( &myx_f, 1, MPI::DOUBLE, 0, msgtags::BOUNDARY_MAX );
                mpi_info.comm.Send( &cpu_time, 1, MPI::DOUBLE, 0, msgtags::EXCHANGE_TIME );
                mpi_info.comm.Send( &my_max_transit, 1, MPI::DOUBLE, 0, msgtags::MAX_TRANSIT );
            }
        }

        {
            /* now we need to broadcast the distribution to everyone. */
            int pDist_array_len = 0;
            double * pDist_array = NULL;
            
            if (mpi_info.arank == 0) {
                /* we promise that we won't write to it.  Really. */
                pDist_array = (double *)pDist->invertedDistribution(pDist_array_len);
            }

            MPI::COMM_WORLD.Bcast(&pDist_array_len, 1, MPI::DOUBLE, mpi_info.aroot);

            if (mpi_info.arank != 0) {
                pDist_array = new double[pDist_array_len];
                memset(pDist_array, 0, pDist_array_len * sizeof(double));
            }

            MPI::COMM_WORLD.Bcast( pDist_array, pDist_array_len, MPI::DOUBLE, mpi_info.aroot );

            if (mpi_info.arank != 0) {
                pDist = new Distribution(pDist_array_len, pDist_array);
                delete[] pDist_array;
            }
        }


        /* ******** find maximum number of supportable processors ********* */
        double max_supp_proc = 0;
        if ( mpi_info.arank >= 0 ) {
            double my_max_supp_proc =
                  (myx_f - myx_i)
                / (1e-300 + container.max_particle_transit(physical_orientation));
            my_max_supp_proc = std::min((double)(INT_MAX-1), my_max_supp_proc);
            log_finest(FILE_LINE ":  mpi_info.comm.Reduce(my_max_supp_proc, &max_supp_proc, 1, MPI::DOUBLE, MPI::SUM, 0);");
            mpi_info.comm.Reduce(&my_max_supp_proc, &max_supp_proc, 1, MPI::DOUBLE, MPI::SUM, 0);
        }
        /* comm.Reduce (above) rooted at aroot; Bcast sent from aroot. */
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&max_supp_proc, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&max_supp_proc, 1, MPI::DOUBLE, mpi_info.aroot);

        /* now that we are done gathering all the information from the last
         * set of active processors, we can create the new active group and
         * make sure that everyone knows who the root of that new group is.
         */
        n_active_proc = std::min(
            mpi_info.Np,
            std::min( int(N / amin_N_per_proc), (int)max_supp_proc)
        );

        /* this is a check, just in case */
        n_active_proc = std::max(1, n_active_proc);

        /* allow some input to limit the number of active procs. */
        if (max_active_proc > 0) {
            n_active_proc = std::min(n_active_proc, max_active_proc);
        }

        if (force_max) n_active_proc = std::min(mpi_info.Np, max_active_proc);

        log_info("new active group contains %d processors", n_active_proc);
        setupActiveComm(n_active_proc);

        /* now calculate the bounds of this node. */
        if ( mpi_info.arank >= 0 ) {
            volb.min[physical_orientation] =
                pDist->leverarm(double(mpi_info.arank    ) / n_active_proc);
            volb.max[physical_orientation] =
                pDist->leverarm(double(mpi_info.arank + 1) / n_active_proc);

            /* we just take the other edges to infinity. */
            volb.min[(physical_orientation+1)%3] = -DBL_MAX;
            volb.min[(physical_orientation+2)%3] = -DBL_MAX;
            volb.max[(physical_orientation+1)%3] = +DBL_MAX;
            volb.max[(physical_orientation+2)%3] = +DBL_MAX;

            log_info(("decompose gives me volb : ("
                        + to_string(V3C(volb.min))
                        + ")-->("
                        + to_string(V3C(volb.max))
                        + ")").c_str());

            /* now copy these edge values over to our boundaries. */
            {
                mpi_boundaries[0].initBoundary(true, physical_orientation, volb.min[physical_orientation]);
                mpi_boundaries[1].initBoundary(false, physical_orientation, volb.max[physical_orientation]);

                mpi_boundaries[0].neighbor = mpi_info.arank -1;
                register int next = mpi_info.arank + 1;
                mpi_boundaries[1].neighbor = next >= n_active_proc ? -1 : next;
            }

            /* make sure that all pointers are removed first because
             * addBoundaryObject does not check for duplicates while
             * removeBoundaryObject DOES remove all instances of a pointer.
             * We do remove-all and then add all because we further remove all
             * edge boundary_ojbects below.  
             */
            container.removeBoundaryObject(&mpi_boundaries[0]);
            container.removeBoundaryObject(&mpi_boundaries[1]);

            container.addBoundaryObject(&mpi_boundaries[0]);
            container.addBoundaryObject(&mpi_boundaries[1]);

            /* fix boundaries for far left and far right. */
            if (mpi_boundaries[0].neighbor < 0) {
                /* I'm on the far left, so min == -DBL_MAX. */
                volb.min[physical_orientation] = -DBL_MAX;
                container.removeBoundaryObject(&mpi_boundaries[0]);
            } else if (mpi_boundaries[1].neighbor < 0) {
                /* I'm on the far right, so min == +DBL_MAX. */
                volb.max[physical_orientation] = +DBL_MAX;
                container.removeBoundaryObject(&mpi_boundaries[1]);
            }

        } else {
            /* this is an inactive node now. */
            setNullVolume();
        }

        /* now sort of this container's _Particle/s into the complete
         * ShuffleQ.
         */
        ShuffleQ * sq = new ShuffleQ[mpi_info.Np];
        for (int i = 0; i < mpi_info.Np; i++) sq[i].neighbor = i;

        sortParticlesIntoSQ(sq, *pDist);

        /* now we do some network intensive work:  send particles to and from
         * all processors.
         */
        sendShuffleQ(sq);

        delete[] sq;

        if ( mpi_info.arank < 0 && container.numberParticles()) {
            THROW(std::runtime_error,   FILE_LINE ":  I'm an inactive node (rank = "
                                      + to_string(mpi_info.rank)
                                      + ") and received "
                                      + to_string(container.numberParticles())
                                      + " Particles?");
        }

        if (mpi_info.arank >= 0) {
            log_info("decompose leaves me with %d particles", container.numberParticles());
        }

        /* now, finally, reset the pid values. */

        lPID.reset();
        rPID.reset();

        delete pDist;
    }/* decomposeEntireDomain */

  private:
    /** send Particle/s to their appropriate new node. */
    inline void sendShuffleQ(ShuffleQ sq[]) {
        /* this is not a time-efficient method of exchanging particles because
         * most of the nodes are waiting during this process. 
         * Oh well.  This is easier to implement. 
         */
        for (int i = 0; i < mpi_info.Np; i++) {
            if (i == mpi_info.rank) {
                /* let's add our particles back into the container. */
                container.addParticles(sq[i].q);
                sq[i].q = NULL;
                log_fine("ShuffleQ logic left me with %d particles", container.numberParticles());
            } else {
                exchangeParticles(MPI::COMM_WORLD, mpi_info.rank, sq[i].neighbor, sq[i].q);
            }
        }/* for */
    }/* sendShuffleQ */

    /** Shuffle all Particle/s into their respective containers.
     * */
    inline void sortParticlesIntoSQ( ShuffleQ sq[], const Distribution & pDist ) {
        /* for each processor, starting with the leftmost, loop through the
         * remaining particles and push them into the shuffle queue. */
        for (int aproc = 0; aproc < n_active_proc; aproc++) {
            double z_max = pDist.leverarm(double(aproc + 1) / n_active_proc);

            /* we start from the back as a trick.  removePtr will always move
             * the last pointer to the removed location.  Thus we don't have
             * to worry about keeping i the same after a positive test. */
            for (int i = container.numberParticles() - 1; i >= 0; i--) {
                _Particle * p = container.getParticles()[i];
                if ( p->pos[physical_orientation] <= z_max ) {
                    /* map aproc # to proc # */
                    int proc = mpi_info.proc_map[aproc];

                    /* add particle to appropriate queue. */
                    p->next = sq[proc].q;
                    sq[proc].q = p;

                    container.removePtr(i);
                }
            }
        }

        /* sometimes the particles do not fit into the comparison done here
         * with <= z_max.  Thus, any remaining particles should be sorted into
         * the last list.
         * As with above, we start from the back as a trick.  removePtr will
         * always move the last pointer to the removed location.  Thus we
         * don't have to worry about keeping i the same after a positive test.
         */
        {
            /* map aproc # to proc # */
            int proc = mpi_info.proc_map[n_active_proc - 1];
            for (int i = container.numberParticles() - 1; i >= 0; i--) {
                _Particle * p = container.getParticles()[i];

                /* add particle to appropriate queue. */
                p->next = sq[proc].q;
                sq[proc].q = p;

                container.removePtr(i);
            }
        }
    }/* sortParticlesIntoSQ */

    /** send particles.
     * This function relies on two template specializations to be provided by
     * for the Particle type:<br>
     * particle_packed_size<Particle>(), which returns the number of bytes
     *     needed to MPI_Pack a Particle;<br>
     * particle_pack<Particle>(uint8_t buffer[], const Particle &), which packs the given
     *     Particle into the given buffer.
     * */
    int sendParticles(MPI::Comm & comm, const int & neighbor, _Particle *& sendlist) {
        /* the buffer that will be used to send _Particle/s. */
        const int ipsz = MPI::INT.Pack_size(1,comm);
        const int max_pack_size
                = ipsz
                + max_packet_length * particle_packed_size<_Particle>(comm);
        uint8_t * buffer = new uint8_t[max_pack_size];

        int total_sent = 0;

        while (sendlist != NULL) {
            int this_packet_length = 0;
            int bpos = 0;
            int tmpInt = 10;

            /* we pack a dummy placeholder in the beginning of the buffer. */
            MPI::INT.Pack((const void*)&tmpInt, 1, (void*)buffer, max_pack_size, bpos, comm);

            while ( this_packet_length < max_packet_length && sendlist != NULL ) {
                _Particle * tmp = sendlist;
                sendlist = sendlist->next;

                /* call the _Particle pack function--
                 * _Particle will have to have corrosponding template
                 * specialization for particle_pack.  The particle_pack
                 * function should return the amount of buffer consumed by
                 * this last packing.
                 */
                particle_pack<_Particle>(buffer, max_pack_size, tmp, bpos, comm);

                /* now get rid of the particle. */
                delete tmp;
                this_packet_length++;
            }/* while */
            
            /* pack the number in this packet into the first part of the
             * buffer. */
            tmpInt = 0;
            MPI::INT.Pack((const void*)&this_packet_length, 1, (void*)buffer, max_pack_size, tmpInt, comm);

            /* now that we have the number of particles packed up, we send
             * them away. */
            log_finer(FILE_LINE ":  sending %d particles", this_packet_length);
            comm.Send(buffer, bpos, MPI::PACKED, neighbor, msgtags::PARTICLES);

            total_sent += this_packet_length;
        }/* while */

        {
            /* now send the DONE NOTICE. */
            int DONE = 0; /* 0 particle size packet indicates DONEness. */
            comm.Send(&DONE, 1, MPI::INT, neighbor, msgtags::PARTICLES);
        }

        /* clean up. */
        delete[] buffer;

        return total_sent;
    }/* sendParticles */

    /** receive particles.
     * This function relies on two template specializations to be provided by
     * for the Particle type:<br>
     * particle_packed_size<Particle>(), which returns the number of bytes
     *     needed to MPI_Pack a Particle;<br>
     * particle_unpack<Particle>(Particle &, const uint8_t buffer[]), which unpacks the given
     *     Particle from the given buffer.
     */
    int receiveParticles(MPI::Comm & comm, const int & neighbor) {
        /* the buffer that will be used to send _Particle/s. */
        const int max_pack_size
                = MPI::INT.Pack_size(1,comm)
                + max_packet_length * particle_packed_size<_Particle>(comm);

        uint8_t * buffer = new uint8_t[max_pack_size];
        int this_packet_length = 0;
        int total_received = 0;
        _Particle * received = NULL;

        /* a temporary storage buffer into which to pack. */
        uint8_t pbuf[sizeof(_Particle)];

        do {
            int bpos = 0;
            /* attempt to receive packet. */
            MPI::Status status;
            comm.Recv(buffer, max_pack_size, MPI::PACKED, neighbor, msgtags::PARTICLES, status);
            log_finer(FILE_LINE ":  received buffer of length %d", status.Get_count(MPI::PACKED));

            /* first unpack the given number of _Particle/s sent. */
            MPI::INT.Unpack((const void*)buffer, max_pack_size, &this_packet_length, 1, bpos, comm);

            total_received += this_packet_length;
            log_finer(FILE_LINE ":  received %d particles", this_packet_length);

            for (int i = 0; i < this_packet_length; i++) {
                particle_unpack(buffer, max_pack_size, (_Particle*)pbuf, bpos, comm);
                _Particle * pptr = new _Particle(*((_Particle*)pbuf));

                if (!volb.contains(pptr->pos)) {
                    log_severe((FILE_LINE "("
                        + to_string(container.getAbsoluteTime())
                        + "s):  particle (from neighbor="
                        + to_string(neighbor)+") at ("
                        + to_string(V3C(pptr->pos)) 
                        + ") ! in volb : ("
                        + to_string(V3C(volb.min))
                        + ")-->("
                        + to_string(V3C(volb.max))
                        + ")").c_str());
                }

                pptr->next = received;
                received = pptr;
            }
        } while (this_packet_length > 0);

        /* now we need to add these particles to our container. */
        container.addParticles(received);

        /* clean up. */
        delete[] buffer;

        return total_received;
    }/* receiveParticles */

    /** Exchange Particle/s with another node. 
     */
    inline void exchangeParticles( MPI::Comm & comm,
                                   const int & myrank,
                                   const int & neighbor,
                                   _Particle *& sendlist) {
        int sent = 0, received = 0;
        if (neighbor < myrank) {
            /* receive first. */
            received = receiveParticles(comm, neighbor);

            /* send second. */
            sent = sendParticles(comm, neighbor,sendlist);
        } else {
            /* send first. */
            sent = sendParticles(comm, neighbor,sendlist);

            /* receive second. */
            received = receiveParticles(comm, neighbor);
        }

        log_finest("exchangeParticles(with neighbor=%d):  sent %d, received %d, have %d",
                 neighbor, sent, received, container.numberParticles());

        if ( mpi_info.arank < 0 && container.numberParticles()) {
            log_severe("Inactive node with particles.  First ten: ");
            for (int i = 0; i < container.numberParticles() && i < 10; i++) {
                _Particle * p = container.getParticles()[i];
                log_severe(("(" + to_string(V3C(p->pos)) + ")").c_str());
            }
        }
    }/* exchangeParticles */

  public:

    /** Perform computational work as provided by container.evolve, and then
     * do MPI-centric-per-timestep operations. */
    void evolve(const double & dt, void * evolve_args = NULL) {
        log_finest("MPI evolve starting with %d particles", container.numberParticles());

        static double shift_err = 0.0;

        {   /* need to allow even non-active containers to "evolve" so that
             * they all keep their timers in sync (if they have such). */
            struct tms ti, tf;
            times(&ti);
            container.evolve(dt, evolve_args);
            times(&tf);

            /* record the amount of time this cycle took.
             * try a running average to minimize wild osillations.
             */
            cpu_time = ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
                         * _for_mpi_seconds_per_clock_tick
                       );
        }

        if (mpi_info.arank >= 0) {
            /* the following loop states that boundaries are always synced
             * from the left to right. */
            shift_err = 0.0;
            for (int i = 0; i < 2; i++) {
                MPIBoundary<_Particle> & bo = mpi_boundaries[i];
                if (bo.neighbor < 0) continue; /* no neighbor. */

                exchangeParticles(mpi_info.comm, mpi_info.arank, bo.neighbor, bo.removed);

                /* now take care of load balancing issues. */
                shift_err += shiftNeighborBoundary(bo);
            }
        }

        double max_supp_proc = 0.0;
        double N = 0;

        /* gather all redecompose test data */
        /* FIXME:  need to change this reduction operation into an
         * asynchronous operation.  There is no reason to force all processors
         * to wait for this barrier. */
        if ( mpi_info.arank >= 0 ) {
            /* find out how many particles there are in the system. */
            double myN = container.numberParticles();
            log_finest(FILE_LINE ":  mpi_info.comm.Reduce(&myN, &N, 1, MPI::DOUBLE, MPI::SUM, 0);");
            mpi_info.comm.Reduce(&myN, &N, 1, MPI::DOUBLE, MPI::SUM, 0);

            /* collect total shift error */
            double new_tot_shift_err = 0.0;
            log_finest(FILE_LINE ":  mpi_info.comm.Reduce(&shift_err, &new_tot_shift_err, 1, MPI::DOUBLE, MPI::SUM, 0);");
            mpi_info.comm.Reduce(&shift_err, &new_tot_shift_err, 1, MPI::DOUBLE, MPI::SUM, 0);

            /* calculate the running average of the shift error */
            tot_shift_err *= shift_err_hfactor;
            tot_shift_err += (1.0 - shift_err_hfactor) * new_tot_shift_err;

            /* calculate the running average of the deviation of the shift
             * error */
            stddev2_tot_shift_err *= shift_err_hfactor;
            stddev2_tot_shift_err += (1.0 - shift_err_hfactor)
                                  * (SQR(new_tot_shift_err) - SQR(tot_shift_err));

            /* collect maximum number of supportable processors. */
            double myx_i = 0, myx_f = 0;
            get_node_edges(myx_i, myx_f);
            double my_max_supp_proc =
                  (myx_f - myx_i)
                / (1e-300 + container.max_particle_transit(physical_orientation));
            my_max_supp_proc = std::min((double)(INT_MAX-1), my_max_supp_proc);
            log_finest(FILE_LINE ":  mpi_info.comm.Reduce(my_max_supp_proc, &max_supp_proc, 1, MPI::DOUBLE, MPI::SUM, 0);");
            mpi_info.comm.Reduce(&my_max_supp_proc, &max_supp_proc, 1, MPI::DOUBLE, MPI::SUM, 0);
        }

        /* now broadcast the piecewise results. */
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&N, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&N, 1, MPI::DOUBLE, mpi_info.aroot);
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&tot_shift_err, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&tot_shift_err, 1, MPI::DOUBLE, mpi_info.aroot);
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&stddev2_tot_shift_err, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&stddev2_tot_shift_err, 1, MPI::DOUBLE, mpi_info.aroot);
        log_finest(FILE_LINE ":  MPI::COMM_WORLD.Bcast(&max_supp_proc, 1, MPI::DOUBLE, aroot);");
        MPI::COMM_WORLD.Bcast(&max_supp_proc, 1, MPI::DOUBLE, mpi_info.aroot);


        /* now let everyone calculate the results. */
        double num_new_procs = ( max_supp_proc - n_active_proc );

        bool need_increase_procs = 
                (/* increase in processors */
                    n_active_proc < mpi_info.Np
                 && ((int)num_new_procs) > 0
                    /* There are roughly too many Particles / proc. */
                 && N >= ( (n_active_proc+1.1) * amin_N_per_proc )
                );

        bool need_decrease_procs =
                (/* decrease in processors */
                    n_active_proc > 1
                 && (
                        ((int)num_new_procs) < 0
                    /* There are roughly not enough Particles / proc.
                     * We'll only worry about removing a proc after
                     * N is (0.5 * amin_N_per_proc) below threshold. */
                     || N <= ( (n_active_proc-0.5) * amin_N_per_proc )
                    )
                );
        
        if ( ( n_active_proc == 1 ||
                /* last load_shift was quite small for the last few
                 * timesteps (this is a running average).  Note that
                 * because of the way I calculated stddev here, it can be
                 * negative.  We'll just fabs() it. */
               sqrt(fabs(stddev2_tot_shift_err)) < load_shift_trigger_frac
             ) &&
             ( need_increase_procs || need_decrease_procs )
           ) {
            log_finest (
                "decompose test: \n"
                " (stddev(load_shift)=%g) < (max(load_shift_err)=%g)\n"
                "\n"
                "need_increase_procs = \n"
                "   (\n"
                "       (n_active_proc=%d) < (mpi_info.Np=%d)\n"
                "    && (int(num_new_procs=%g)) > 0\n"
                "    && (N=%g) >= ( ((n_active_proc=%d) + 1) * (amin_N_per_proc=%g) )\n"
                "   ) == %s;\n"
                "\n"
                "need_decrease_procs = \n"
                "   (\n"
                "       (n_active_proc=%d) > 1\n"
                "    && (\n"
                "           (int(num_new_procs=%g)) < 0\n"
                "        || (N=%g) <= ( ((n_active_proc=%d) - 0.5) * (amin_N_per_proc=%g) )\n"
                "       )\n"
                "   ) == %s;\n"
                , sqrt(fabs(stddev2_tot_shift_err)), load_shift_trigger_frac,
                n_active_proc, mpi_info.Np,
                num_new_procs, 
                N, n_active_proc, amin_N_per_proc,
                (need_increase_procs ? "true" : "false"),
                n_active_proc,
                num_new_procs,
                N, n_active_proc, amin_N_per_proc,
                (need_decrease_procs ? "true" : "false")
            );
            decomposeEntireDomain();
        }

        /* done! */
        log_finest("MPI evolve ending with %d particles", container.numberParticles());
    }

    /** The bounded volume of this MPI node. */
    octree::vol_bounds volb;

    /** Add Particles to the MPI containers.
     * Note that it is best designed to add all particles to all active
     * nodes (for portability throughout the dynamic load balancing).
     * This function will silently remove from the list <em>all</em> particles
     * that do not reside within the nodal boundaries.
     *
     * @param input
     *     A linked list of Particle/s.
     */
    inline int addParticles(_Particle * input) {
        /* trim list */
        _Particle ** lptr = &input;
        while (*lptr) {
            if ( !volb.contains((*lptr)->pos) ) {
                _Particle * tmp = *lptr;
                *lptr = (*lptr)->next;
                delete tmp;
            } else lptr = &(*lptr)->next;
        }

        return container.addParticles(input);
    }


    /** Split particles into smaller pieces.  This may result in particles
     * that are not exactly in boundaries.
     */
    inline int splitParticles(const double & old_size,
                              const double & new_size,
                              const Vector<double,3> & density = DBL_MAX ) {
        return container.splitParticles(old_size,new_size,density);
    }

    /** whether initialized. */
    bool initialized;

    /** 1-D specialization:  The physical direction of the 1-D array.
     * ?TODO?:  Need to fix this so we can do 2-D and even 3-D load-balancing.
     * I'm not sure what 4-D would be useful for.
     */
    int physical_orientation;

    /** Storage container for MPI information (rank, communicator, ...). */
    struct {
        /** Number of processes. */
        int Np;

        /** Communicator for active grid. */
        MPI::Intracomm comm;

        /** Rank in the active COMM.
         * This will be > 0 for an inactive node.
         */
        int arank;

        /** Rank of process in COMM_WORLD. */
        int rank;

        /** Rank of the root of active grid communications (arank == 0) with
         * respect to the MPI::COMM_WORLD communicator.
         */
        int aroot;

        /** mapping of active processor ranks to WORLD ranks. */
        int * proc_map;
    } mpi_info;

    /** Print the header-row for the data printed by printInfo(). */
    std::ostream & printInfoHeader(std::ostream & output) const {
        output << "max[X..Z] min[X..Z] container_max[X..Z] container_min[X..Z] cpu_time "
                  "load_balance_lshift load_balance_rshift numberParticles max_transit "
                  "tot_shift_err stddev_tot_shift_err";
        return output;
    }

    /** Print some various bits of node data that are useful when
     * debugging/profiling the MPI-centric code. */
    std::ostream & printInfo(std::ostream & output) const {
        volb.print(output) <<  '\t';

        if (container.numberParticles() > 1) {
            octree::vol_bounds cvolb;
            container.measureCurrentRcmAndVol(cvolb);
            cvolb.print(output);
        } else {
            output << "0\t0\t0\t0\t0\t0";
        }

        output << '\t' << cpu_time
               << '\t' << load_balance_lshift
               << '\t' << load_balance_rshift
               << '\t' << container.numberParticles()
               << '\t' << container.max_particle_transit(physical_orientation)
               << '\t' << tot_shift_err
               << '\t' << sqrt(fabs(stddev2_tot_shift_err));
        return output;
    }

    /** Check all the particles in the container to see if they are still
     * within the compute node's domain. */
    int check_particles() const {
        int nout = 0;
        for (int i = 0; i < container.numberParticles(); i++) {
            nout += !volb.contains(container.getParticles()[i]->pos);
        }
        return nout;
    }

    /** The actual Particle container underneath these nodes. */
    ParticleContainer container;

    /** Execution time information. */
    double cpu_time;

    /** Amount of time spent waiting for neighbors to catch up. */
    double wait_time;

    /** Boundaries. */
    MPIBoundary<_Particle> mpi_boundaries[2];

    /** Average minimum number of particles per processor. */
    double amin_N_per_proc;

    /** Maximum number of Particles to send in one packet. */
    int max_packet_length;

    /** A record of how much the last load balance shift changed the
     * boundaries on the left. */
    double load_balance_lshift;

    /** A record of how much the last load balance shift changed the
     * boundaries on the right. */
    double load_balance_rshift;

    /** Number of active processors. */
    int n_active_proc;

    /** The maximum fraction of the load_shift to volume size that is needed
     * to perform re-decomposition of the entire computational domain.
     */ 
    double load_shift_trigger_frac;

    /** The error historisis.  This is compared to load_shift_trigger_frac.
     * This is only used by aroot. */
    double tot_shift_err;

    /** The std. dev. of error historisis. */
    double stddev2_tot_shift_err;

    /** Historisis factor for tot_shift_err.  It is used as:
     * tot_shift_err =
     *     old_tot_shift_err*shift_err_hfactor
     *   + (1-shift_err_factor)*new_tot_shift_err. 
     *
     * This must be between 0 and 1.
     */
    double shift_err_hfactor;

    /** The feedback parameters to use for load balancing.
     * Be careful.  The semantics used here are not rigorously correct to
     * similar electrical control systems.
     *
     * The shift in position \f$ d\mathfrak{L}_{\rm SHIFT} \f$ of an MPI
     * boundary between two neighboring MPI * nodes is calculated as:
     \f[
        d\mathfrak{L}_{\rm SHIFT}
          = 
            \frac{1}{2}
            \frac{\mathfrak{L}_1 + \mathfrak{L}_2}
                 {\left(\tau_1 + \tau_2 \right)}
           \cdot 
            \left(
                  {\rm K}_{\rm P} \epsilon
                + \frac{1}{{\rm T}_{\rm i}} \int_{0}^{t} \epsilon\ dt
                - {\rm T}_{\rm D}\frac{\partial\epsilon}{\partial t}
            \right)
     \f]
     */
    class PID {
      public: 
        /** Controller constructor. */
        PID (const double & kp =  0.1,
             const double & ti =  100.0,
             const double & td =  1.0/80.0 ) :
            Kp(kp),
            Ti(ti),
            Td(td),
            last_error(0.0),
            error_integral(0.0) {}

        /** Proportional gain constant. */
        double Kp;
        /** Integral gain constant. */
        double Ti;
        /** Differential gain constant. */
        double Td;

        /** Accumulate the integral error and return the result. */
        inline const double & IntegralError(const double & next_error) {
            error_integral += next_error;
            return error_integral;
        }

        /** Calculate the deviation from the last error to the current error
         * and return the result. 
         * This also saves the current error for the next call to this
         * function.
         */
        inline double dE(const double & next_error) {
            register double retval = next_error - last_error;
            last_error = next_error;
            return retval;
        }

        /** Reset the controller.
         * Reset amounts to zeroing the integral error accumulator used with
         * integralError(const double &) and the last_error register for used
         * for dE(const double &).
         */
        inline void reset() {
            error_integral = 0.0;
            last_error = 0.0;
        }

      private:
        double last_error;
        double error_integral;
    };

    /** The PID controller for the interface with the neighbor on the right
     * side of this MPI node. */
    PID rPID;

    /** The PID controller for the interface with the neighbor on the left
     * side of this MPI node. */
    PID lPID;

    /** Boolean value to specify whether each MPI node should attempt to
     * create its own working subdirectory based on its MPI rank. 
     * This is useful when you have multiple MPI nodes creating separate
     * output files.  
     * [Default true].
     */
    bool wdir_change;

    /** Mode of created working subdirectory.
     * [Default 
     * \code S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH \endcode
     * ]
     */
    mode_t wdir_perm;

    /** Boolean value to specify whether we attempt to dump core when sigterm
     * is received.  SIGTERM is used by PBS often to halt jobs.  [Default
     * true] */
    bool sigterm_core_dumps;
};

#endif // PARTICLENODE_H
