// -*- c++ -*-
// $Id: ParticleSrc.h,v 1.3 2005/06/07 10:00:16 olsonse Exp $
/*
 * Copyright 2002-2004 Spencer Olson
 *
 * $Log: ParticleSrc.h,v $
 * Revision 1.3  2005/06/07 10:00:16  olsonse
 * Many changes over the course of time
 *
 * Revision 1.2  2005/01/19 11:45:43  olsonse
 * Small fixes.
 *
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 * Revision 1.6  2004/11/29 14:48:34  olsonse
 * Lots of changes,  check diffs
 *
 * Revision 1.5  2004/04/13 15:54:08  labrat
 * Simplified and cleaned up gauss_deviate which is used to create the input
 * beam.
 * gauss_deviate now accepts only one parameter (sigma of distribution).
 *
 * Revision 1.4  2004/04/09 12:32:35  olsonse
 * small changes.
 *
 * Revision 1.3  2003/08/07 09:18:41  olsonse
 * Many updates and improvments on samplers and so on.
 * Pulses are now of some duration.
 *
 * Revision 1.2  2003/07/14 21:16:21  olsonse
 * Commiting a version that seems to work.  Getting ready to include real field
 * calculations preparatory to evaporation work as well as modelling the verticle
 * rise.
 *
 * Revision 1.1.1.1  2003/05/18 02:05:10  olsonse
 * Initial import of atomguide into cvs.
 * This marks the change from a running version that did not really do cells
 * correctly, to one where cells will correctly be done (hopefully).
 *
 */

#ifndef PARTICLESRC_H
#define PARTICLESRC_H

#include "Particle.h"
#include <olson-tools/Distribution.h>
#include <olson-tools/random/random.h>
#include <olson-tools/physical.h>
#include <olson-tools/Vector.h>

/** A <em>very</em> simple xyz position distribution: pointsource.
*/
class PointSource {
  public:
    /** Not too much to initialize, but this is required by the position
     * distribution interface of ParticleSrc<...>.
     */
    void initialize(const Vector<double,3> & origin) { }
    inline double x() const { return 0.0; }
    inline double y() const { return 0.0; }
    inline double z() const { return 0.0; }
};

class Gaussian3DParticleDistrib {
  public:
    Gaussian3DParticleDistrib() {
        sigma_x = (1.5*physical::unit::cm);     /* default to 1.5cm */
        max_x = 4;
        min_x = -4;
        nbins = 1000;
        memset (distro, 0, sizeof (distro));
    }

    ~Gaussian3DParticleDistrib() {
        freeDistro();
    }

    void initialize(const Vector<double,3> & origin) {
        freeDistro();

        /* beta = 0.5 / sigma^2 */
        Vector<double,3> sigma2 = compMult(sigma_x, sigma_x);
        Vector<double,3> beta = 0.5;
        beta.compDiv(sigma2);

        distro[X] = new Distribution(GaussianDistrib(beta[X]), min_x[X]*sigma_x[X], max_x[X]*sigma_x[X], nbins[X]);
        distro[Y] = new Distribution(GaussianDistrib(beta[Y]), min_x[Y]*sigma_x[Y], max_x[Y]*sigma_x[Y], nbins[Y]);
        distro[Z] = new Distribution(GaussianDistrib(beta[Z]), min_x[Z]*sigma_x[Z], max_x[Z]*sigma_x[Z], nbins[Z]);
    }/* initialize */

    inline double x() const {
        return distro[X]->lever();;
    }

    inline double y() const {
        return distro[Y]->lever();
    }

    inline double z() const {
        return distro[Z]->lever();
    }

    Vector<double,3>	sigma_x;
    Vector<double,3>    min_x;
    Vector<double,3>    max_x;
    Vector<int,3>       nbins;

  private:
    Distribution * distro[3];

    inline void freeDistro() {
        for (int i = 0; i < 3; i++) {
            if (distro[i]) delete distro[i];
            distro[i] = NULL;
        }
    }
};

class FlatParticleDistrib {
  public:
    FlatParticleDistrib() {
        delta_x = (1.5*physical::unit::cm);     /* default to 1.5cm */
    }

    void initialize(const Vector<double,3> & origin) {
    }/* initialize */

    inline double x() const {
        return ((MTRNGrand()*2.0 - 1.0) * delta_x[X]);
    }

    inline double y() const {
        return ((MTRNGrand()*2.0 - 1.0) * delta_x[Y]);
    }

    inline double z() const {
        return ((MTRNGrand()*2.0 - 1.0) * delta_x[Z]);
    }

    Vector<double,3>	delta_x;
};

class Gaussian3Distros {
  public:
    typedef GaussianDistrib P_vx;
    typedef GaussianDistrib P_vy;
    typedef GaussianDistrib P_vz;
};

template <class SpatialDistribution = Gaussian3DParticleDistrib,
          class VDistros = Gaussian3Distros,
          class _Particle = Particle>
class ParticleSrc {
  public:
    inline ParticleSrc(const ParticleSrc & that) {
        memset (vdistro, 0, sizeof (vdistro));
        *this = that;
    }

    inline ParticleSrc() :
        pos_distrib() {
        memset (vdistro, 0, sizeof (vdistro));
        set_defaults();
    }

    template <class T>
    inline ParticleSrc(const T & dist_init) :
        pos_distrib(dist_init) {
        memset (vdistro, 0, sizeof (vdistro));
        set_defaults();
    }

    inline ~ParticleSrc() {
        freevDistro();
    }
    
    /* create the appropriate number and placement of atoms for the given dt
     * */
    _Particle * createParticles(const double & dt);

    inline void changeFlux(const double & flx) {
        flux = flx;
        normalized_flux = flux / seed.size();
    }

    /** Initialize the class for work.
     * This must be done AFTER the parameters are chosen, and
     * BEFORE any atoms can be inserted into the module.
     * */
    void initialize() {
        changeFlux(flux);

        /* vel_distrib.initialize(); */
        /* the sigma_v? parameter velocity distribution is used as:
         * p(v) = 1/Z * exp(- v^2 / ( 2 kB T / m) )
         *      = 1/Z * exp(- v^2 / (2 * sigma_v?^2 )
         *
         * remember that 1/2 m sigma(v?)^2 = 1/2 kB T?
         * and
         * beta = 1/2 m / (kB T)
         *
         * for the Distribution class, we'll use beta whereas for gauss_deviate we
         * use sigma_v
         * */
        Vector<double,3> sigma_v;
        sigma_v = temperature
                * (physical::constant::K_B/seed.mass());

        Vector<double,3> beta = 0.5;
        beta.compDiv(sigma_v); /* this is really sigma_v^2 */

        sigma_v.save_sqrt();

        freevDistro();
        vdistro[X] = new Distribution(typename VDistros::P_vx(beta[X],velocity[X]), v_range.min[X]*sigma_v[X], v_range.max[X]*sigma_v[X], nbins[X]);
        vdistro[Y] = new Distribution(typename VDistros::P_vy(beta[Y],velocity[Y]), v_range.min[Y]*sigma_v[Y], v_range.max[Y]*sigma_v[Y], nbins[Y]);
        vdistro[Z] = new Distribution(typename VDistros::P_vz(beta[Z],velocity[Z]), v_range.min[Z]*sigma_v[Z], v_range.max[Z]*sigma_v[Z], nbins[Z]);

        /* we pass in the origin even though we add it on later because some
         * position distributions may be in the business of inverting
         * distributions.  If this is the case, these distributions MUST know
         * exactly where they are to be evaluated.
         */
        pos_distrib.initialize(origin);
    }

    /** gather statisical data from modules.
     * */
    void reportStatistics(std::ostream & output) const;

    inline const ParticleSrc & operator=(const ParticleSrc & that) {
        pos_distrib             = that.pos_distrib;
        flux                    = that.flux;
        origin                  = that.origin;
        temperature             = that.temperature;
        velocity                = that.velocity;

        pulsed_source           = that.pulsed_source;
        pulse_duration          = that.pulse_duration;
        pulse_timer             = that.pulse_timer;

        seed                    = that.seed;

        normalized_flux         = that.normalized_flux;
        running_total_flux      = that.running_total_flux;
        total_time              = that.total_time;

        v_range.min             = that.v_range.min;
        v_range.max             = that.v_range.max;

        nbins                   = that.nbins;

        freevDistro();
        for (int i = 0; i < 3; i++) {
            vdistro[i] = new Distribution(*that.vdistro[i]);
        }

        return *this;
    }

    /** Get a pointer to the particular velocity distribution for a certain
     * direction.
     * The caller should check to see that the return value is not NULL.
     */
    inline Distribution * getVDistrib(const int & i) {
        return vdistro[i];
    }

  private:
    inline void freevDistro() {
        for (int i = 0; i < 3; i++) {
            if (vdistro[i]) delete vdistro[i];
            vdistro[i] = NULL;
        }
    }

  public:
    SpatialDistribution pos_distrib;
    double	       flux; /* Flux in atoms/s, fixed to a given multiple of ATOMS_PER_MACROATOM */
    Vector<double,3>   origin;
    Vector<double,3>   temperature;

    /** Average velocity of atoms.
     * @defaults 0 m/s.
     */
    Vector<double,3>	velocity;
    bool	pulsed_source;
    double	pulse_duration;
    double	pulse_timer;

    /** The range of the velocity distribution inversion (in units of
     * sigma_v[x,y,z]).
     * min and max == 4.0 is good for gaussians.
     * min == 0.0 and max = 5.0 is good for 2D maxwellians.
     * 
     * Defaults:  min = -4, max = 4.
     * */
    struct {
        Vector<double,3> min;
        Vector<double,3> max;
    } v_range;

    /** The number of bins to use for velocity distributions. */
    Vector<int,3> nbins;

    _Particle seed;

  private:
    void        set_defaults ();
    double      normalized_flux;
    double      running_total_flux;
    double	total_time;
    Distribution * vdistro[3];
};

#include "ParticleSrc.C"

#endif // PARTICLESRC_H
