// -*- c++ -*-
// $Id: Wall.C,v 1.1.1.1 2005/01/08 04:57:25 olsonse Exp $
/*
 * Copyright 2002-2004 Spencer Olson
 *
 * $Log: Wall.C,v $
 * Revision 1.1.1.1  2005/01/08 04:57:25  olsonse
 * Initial CVS import (finally).
 *
 *
 */

#include "Wall.h"
#include <olson-tools/logger.h>
#include <olson-tools/power.h>
#include <olson-tools/ompexcept.h>
#include <math.h>
#include <stdexcept>

using namespace physical::constants;
using namespace physical::units;

inline Wall::wall_distrib::wall_distrib(const double & temperature,
                                        const int & particle_type,
                                        const double & particle_mass,
                                        const int & nbins) :
    beta(0.5 * particle_mass / (K_B * temperature)),
    sigma_v( sqrt( 0.5 / beta) ),

    maxwellian(
        MaxwellianDistrib2D(beta),
        /* range is from 0 to 5 \sigma. */
        0.0,
        5.0*sigma_v,
        nbins),

    gaussian(
        GaussianDistrib(beta),
        /* range is from -4 \sigma to 4 \sigma. */
        -4.0*sigma_v,
        4.0*sigma_v,
        nbins)
{ }

void Wall::addDistrib( const int & particle_type, const double & particle_mass ) {
    /* first make sure there are enough bins. */
    if (distrib_length <= particle_type) {
        /* we use realloc because it makes it easier to combine allocation
         * with copying over old values. */
        register int nlen = particle_type + 1;
        distrib = (wall_distrib**) realloc(distrib, sizeof(void*) * nlen);
        /* initialize any new elements to NULL. */
        memset(&distrib[distrib_length], 0, sizeof(void*)* (nlen - distrib_length));
        distrib_length = nlen;
    }

    if (distrib[particle_type] != NULL) {
        THROW(std::runtime_error,"attempting to overwrite energy distribution for particle type");
    }

    distrib[particle_type] = new wall_distrib(temperature, particle_type, particle_mass, nbin_distrib);
}

void Wall::setSurfaceMatrices() {
    /* now we are ready to create the reflection matrix.
     * I know that the method I am doing for this is not the fastest, but who
     * cares.  It is the most transparent and I only have to do this once
     * anyway!
     * */
    /* cos(phi) = x/sqrt(x**2 + y**2)
     * sin(phi) = y/sqrt(x**2 + y**2)
     * cos(theta (polar)) = z/r
     * sin(theta) = sqrt(1-cos(theta)**2)
     * */
    double costheta = n[Z] /* /r */;
    double sintheta = sqrt(1 - SQR(costheta));
    double cosphi = 1, sinphi = 0;
    if (sintheta) {
        cosphi = n[X] / hypot( n[X], n[Y]);
        sinphi = n[Y] / hypot( n[X], n[Y]);
    }

    /* our reflection matrix is given by:
     * Rz[-phi].Ry[-theta].ReflectVz.Ry[theta].Rz[phi]
     * where Rz[x] and Ry[x] are passive rotations.
     * */
    const double rz_[3][3]  =
                         {{ cosphi,    sinphi,    0   },
                          {-sinphi,    cosphi,    0   },
                          {   0,         0,       1   }};

    const double ry_[3][3]  =
                         {{ costheta,    0,  -sintheta},
                          {   0,         1,       0   },
                          { sintheta,    0,   costheta}};

    const double vrefl_[3][3]  = 
                         {{   1,         0,       0   },
                          {   0,         1,       0   },
                          {   0,         0,      -1   }};

    const double ry2_[3][3]  =
                         {{ costheta,    0,   sintheta},
                          {   0,         1,       0   },
                          {-sintheta,    0,   costheta}};

    const double rz2_[3][3]  =
                         {{ cosphi,   -sinphi,    0   },
                          { sinphi,    cosphi,    0   },
                          {   0,         0,       1   }};

    SquareMatrix<double,3> & rz    = *((SquareMatrix<double,3>*)rz_);
    SquareMatrix<double,3> & ry    = *((SquareMatrix<double,3>*)ry_);
    SquareMatrix<double,3> & vrefl = *((SquareMatrix<double,3>*)vrefl_);
    SquareMatrix<double,3> & ry2   = *((SquareMatrix<double,3>*)ry2_);
    SquareMatrix<double,3> & rz2   = *((SquareMatrix<double,3>*)rz2_);

    M_r = rz2 * ry2 * vrefl * ry * rz;
    R   = rz2 * ry2;
}

